package com.nesterovskyBros.annotation.processor;

import java.util.HashSet;
import java.util.ServiceLoader;
import java.util.Set;

import javax.annotation.processing.AbstractProcessor;
import javax.annotation.processing.RoundEnvironment;
import javax.annotation.processing.SupportedOptions;
import javax.annotation.processing.SupportedSourceVersion;

import javax.lang.model.SourceVersion;

import javax.lang.model.element.Element;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.TypeElement;

import com.nesterovskyBros.annotation.Yield;

/**
 * A post compile time processor to refactor yield methods.
 */
@SupportedSourceVersion(value = SourceVersion.RELEASE_7)
@SupportedOptions("debug")
public class YieldProcessor extends AbstractProcessor
{
  /**
   * Returns the names of the annotation types supported by this processor.
   * @return the names of the annotation types supported by this processor.
   */
  public Set<String> getSupportedAnnotationTypes() 
  {
    HashSet<String> set = new HashSet<String>(1);
    
    set.add(Yield.class.getName());
    
    return set;
  }

  /**
   * {@inheritDoc}
   */
  public boolean process(
    Set<? extends TypeElement> annotations,
    RoundEnvironment roundEnv)
  {
    HashSet<ExecutableElement> methods = new HashSet<ExecutableElement>();

    for(TypeElement type: annotations)
    {
      for(Element element: roundEnv.getElementsAnnotatedWith(type))
      {
        if (element instanceof ExecutableElement)
        {
          ExecutableElement method = (ExecutableElement)element;
  
          methods.add(method);
        }
      }
    }
    
    if (methods.isEmpty())
    {
      return false;
    }
    
    for(YieldService service: ServiceLoader.load(YieldService.class))
    {
      if (service.supports(processingEnv, roundEnv))
      {
        return service.transform(processingEnv, roundEnv, methods);
      }
    }
    
    return false;
  }
}
