package com.nesterovskyBros.annotation.processor;

import java.util.Set;

import javax.annotation.processing.ProcessingEnvironment;
import javax.annotation.processing.RoundEnvironment;

import javax.lang.model.element.ExecutableElement;

/**
 * A service to support yield transformation for different compilers.
 */
public interface YieldService 
{
  /**
   * Checks whether transformation is 
   * supported for this processing and round environments. 
   * @param processingEnv a {@link ProcessingEnvironment} instance.
   * @param roundEnv a {@link RoundEnvironment} instance.
   * @return true if transformation is supported, and false otherwise.
   */
  boolean supports(
    ProcessingEnvironment processingEnv,
    RoundEnvironment roundEnv);
  
  /**
   * Transforms methods.
   * @param methods methods to transform.
   * @param processingEnv a {@link ProcessingEnvironment} instance.
   * @param roundEnv a {@link RoundEnvironment} instance.
   * @return true if operation succeeds, and false otherwise.
   */
  boolean transform(
    ProcessingEnvironment processingEnv,
    RoundEnvironment roundEnv,
    Set<ExecutableElement> methods);
}
