package com.nesterovskyBros.annotation.processor.eclipse;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.annotation.processing.ProcessingEnvironment;
import javax.annotation.processing.RoundEnvironment;

import javax.lang.model.element.ExecutableElement;
import javax.tools.Diagnostic.Kind;

import org.eclipse.jdt.internal.compiler.ASTVisitor;

import org.eclipse.jdt.internal.compiler.apt.dispatch.BaseProcessingEnvImpl;
import org.eclipse.jdt.internal.compiler.apt.dispatch.RoundEnvImpl;

import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.ast.MethodDeclaration;

import org.eclipse.jdt.internal.compiler.lookup.ClassScope;

import com.nesterovskyBros.annotation.processor.eclipse.YieldTransformer;

/**
 * A transformer factory for the yield processor.
 */
public class YieldService
  implements com.nesterovskyBros.annotation.processor.YieldService 
{
  /**
   * {@inheritDoc}
   */
  public boolean supports(
    ProcessingEnvironment processingEnv,
    RoundEnvironment roundEnv)
  {
    try
    {
      Class<?> type = Class.forName(
        "org.eclipse.jdt.internal.compiler.apt.dispatch.BaseProcessingEnvImpl");
    
      return type.cast(processingEnv) != null;
    }
    catch(Throwable e)
    {
      return false;
    }
  }
  
  /**
   * {@inheritDoc}
   */
  public boolean transform(
    ProcessingEnvironment processingEnv,
    RoundEnvironment roundEnv,
    final Set<ExecutableElement> methods)
  {
    
    return Runner.transform(processingEnv, roundEnv, methods);
  }
  
  private static class Runner
  {
    public static boolean transform(
      ProcessingEnvironment processingEnv,
      RoundEnvironment roundEnv,
      final Set<ExecutableElement> methods)
    {
      final BaseProcessingEnvImpl environment = 
        (BaseProcessingEnvImpl)processingEnv;
      
      RoundEnvImpl roundEnvironment = (RoundEnvImpl)roundEnv;
      CompilationUnitDeclaration[] units;
      
      try
      {
        Field field = roundEnvironment.getClass().getDeclaredField("_units");
        boolean accessible = field.isAccessible();
        
        field.setAccessible(true);
        
        try
        {
          units = (CompilationUnitDeclaration[])field.get(roundEnvironment);
        }
        finally
        {
          field.setAccessible(accessible);
        }
      }
      catch(Exception e)
      {
        return false;
      }
      
      final HashMap<ExecutableElement, MethodDeclaration> declarations = 
        new HashMap<ExecutableElement, MethodDeclaration>(); 

      ASTVisitor visitor = new ASTVisitor()
      {
        public boolean visit(
          MethodDeclaration methodDeclaration, 
          ClassScope scope) 
        {
          ExecutableElement element = (ExecutableElement)environment.getFactory().
            newElement(methodDeclaration.binding);

          if (methods.contains(element))
          {
            declarations.put(element, methodDeclaration);
          }
          
          return true;
        }
      };
      
      for(CompilationUnitDeclaration unit: units)
      {
        int count = declarations.size();
        
        unit.traverse(visitor, unit.scope);
        
        // Populate method bodies.
        if (count < declarations.size())
        {
          environment.getCompiler().parser.getMethodBodies(unit);
        }
      }
      
      boolean debug = environment.getOptions().containsKey("debug");
      boolean succeed = false;
      
      for(Map.Entry<ExecutableElement, MethodDeclaration> entry: 
        declarations.entrySet())
      {
        YieldTransformer transformer = 
          new YieldTransformer(environment, entry.getKey(), entry.getValue());
        
        succeed |= transformer.transform();
        
        if (debug)
        {
          environment.getMessager().printMessage(
            Kind.NOTE, 
            entry.getValue().toString());
        }
      }
      
      return succeed;
    }
  }
}
