package com.nesterovskyBros.annotation.processor.eclipse;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import javax.lang.model.element.Element;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.util.Elements;
import javax.lang.model.util.Types;

import javax.tools.Diagnostic.Kind;

import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.apt.dispatch.BaseProcessingEnvImpl;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.AllocationExpression;
import org.eclipse.jdt.internal.compiler.ast.Annotation;
import org.eclipse.jdt.internal.compiler.ast.Argument;
import org.eclipse.jdt.internal.compiler.ast.ArrayAllocationExpression;
import org.eclipse.jdt.internal.compiler.ast.ArrayInitializer;
import org.eclipse.jdt.internal.compiler.ast.ArrayQualifiedTypeReference;
import org.eclipse.jdt.internal.compiler.ast.AssertStatement;
import org.eclipse.jdt.internal.compiler.ast.Assignment;
import org.eclipse.jdt.internal.compiler.ast.BinaryExpression;
import org.eclipse.jdt.internal.compiler.ast.Block;
import org.eclipse.jdt.internal.compiler.ast.BreakStatement;
import org.eclipse.jdt.internal.compiler.ast.CaseStatement;
import org.eclipse.jdt.internal.compiler.ast.CastExpression;
import org.eclipse.jdt.internal.compiler.ast.ConstructorDeclaration;
import org.eclipse.jdt.internal.compiler.ast.ContinueStatement;
import org.eclipse.jdt.internal.compiler.ast.DoStatement;
import org.eclipse.jdt.internal.compiler.ast.EmptyStatement;
import org.eclipse.jdt.internal.compiler.ast.EqualExpression;
import org.eclipse.jdt.internal.compiler.ast.ExplicitConstructorCall;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.FalseLiteral;
import org.eclipse.jdt.internal.compiler.ast.FieldDeclaration;
import org.eclipse.jdt.internal.compiler.ast.ForStatement;
import org.eclipse.jdt.internal.compiler.ast.ForeachStatement;
import org.eclipse.jdt.internal.compiler.ast.IfStatement;
import org.eclipse.jdt.internal.compiler.ast.InstanceOfExpression;
import org.eclipse.jdt.internal.compiler.ast.IntLiteral;
import org.eclipse.jdt.internal.compiler.ast.LabeledStatement;
import org.eclipse.jdt.internal.compiler.ast.Literal;
import org.eclipse.jdt.internal.compiler.ast.LocalDeclaration;
import org.eclipse.jdt.internal.compiler.ast.MessageSend;
import org.eclipse.jdt.internal.compiler.ast.MethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.NullLiteral;
import org.eclipse.jdt.internal.compiler.ast.OperatorIds;
import org.eclipse.jdt.internal.compiler.ast.
  ParameterizedQualifiedTypeReference;
import org.eclipse.jdt.internal.compiler.ast.ParameterizedSingleTypeReference;
import org.eclipse.jdt.internal.compiler.ast.PostfixExpression;
import org.eclipse.jdt.internal.compiler.ast.QualifiedAllocationExpression;
import org.eclipse.jdt.internal.compiler.ast.QualifiedNameReference;
import org.eclipse.jdt.internal.compiler.ast.QualifiedTypeReference;
import org.eclipse.jdt.internal.compiler.ast.Reference;
import org.eclipse.jdt.internal.compiler.ast.ReturnStatement;
import org.eclipse.jdt.internal.compiler.ast.SingleMemberAnnotation;
import org.eclipse.jdt.internal.compiler.ast.SingleNameReference;
import org.eclipse.jdt.internal.compiler.ast.SingleTypeReference;
import org.eclipse.jdt.internal.compiler.ast.Statement;
import org.eclipse.jdt.internal.compiler.ast.StringLiteral;
import org.eclipse.jdt.internal.compiler.ast.SuperReference;
import org.eclipse.jdt.internal.compiler.ast.SwitchStatement;
import org.eclipse.jdt.internal.compiler.ast.SynchronizedStatement;
import org.eclipse.jdt.internal.compiler.ast.ThisReference;
import org.eclipse.jdt.internal.compiler.ast.ThrowStatement;
import org.eclipse.jdt.internal.compiler.ast.TrueLiteral;
import org.eclipse.jdt.internal.compiler.ast.TryStatement;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.ast.UnaryExpression;
import org.eclipse.jdt.internal.compiler.ast.WhileStatement;
import org.eclipse.jdt.internal.compiler.classfmt.ClassFileConstants;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.TypeIds;

/**
 * A transformer for the yield processor.
 */
public class YieldTransformer
{
  /**
   * Creates a {@link YieldTransformer} instance.
   * @param environment a processing environment.
   * @param roundEnvironment a round environment.
   * @param element a method element.
   * @param method a method element.
   * @param declaration a method declaration.
   */
  public YieldTransformer(
    BaseProcessingEnvImpl environment,
    ExecutableElement method, 
    MethodDeclaration declaration)
  {
    this.environment = environment;
    this.method = method;
    this.declaration = declaration;
    this.types = environment.getTypeUtils();
    this.elements = environment.getElementUtils();
  }

  /**
   * Transforms the method body.
   * @return true if operation succeeds, and false otherwise.
   */
  public boolean transform()
  {
    try
    {
      if (!validate())
      {
        return false;
      }
    }
    catch(Exception e) 
    {
      return false;
    }
    
    return refactor();
  }
  
  /**
   * Issue an error and throws exception. 
   * @param message an error message.
   */
  private void invalid(String message)
  {
    environment.getMessager().printMessage(Kind.ERROR, message, method);
    
    throw new IllegalStateException(message);
  }
  
  
  /**
   * Validates assumptions.
   * @return true if refactoring is possible, and false otherwise.
   */
  private boolean validate()
  {
    if ((declaration.modifiers &  ClassFileConstants.AccSynchronized) != 0)
    {
      invalid("Method should be not synchronized.");
    }

    if (declaration.isAbstract())
    {
      invalid("Method should be not abstract.");
    }
    
    if (declaration.arguments != null)
    {
      for(Argument argument: declaration.arguments)
      {
        if ((argument.modifiers & ClassFileConstants.AccFinal) == 0)
        {
          invalid("Parameters should be final.");
        }
      }
    }
    
    Element type = types.asElement(method.getReturnType());
    
    if (type.equals(elements.getTypeElement("java.lang.Iterable")))
    {
      iterable = true;
    }
    else if (type.equals(elements.getTypeElement("java.util.Iterator")))
    {
      iterable = false;
    }
    else
    {
      invalid("Return type should be either Iterable or Iterator.");
    }
    
    if (declaration.statements == null)
    {
      invalid("Method body is required.");
    }
    
    for(Statement statement: declaration.statements)
    {
      if ((statement instanceof TypeDeclaration) ||
        (statement instanceof EmptyStatement))
      {
        continue;
      }

      if (statement instanceof LocalDeclaration)
      {
        iteratorDecl = (LocalDeclaration)statement;
        
        if (!isArrayList(iteratorDecl.type) || 
          !isNewArrayList(iteratorDecl.initialization))
        {
          iteratorDecl = null;
        }
      }

      break;
    }
    
    if (iteratorDecl == null)
    {
      invalid("The first statement should be of the form " +
        "ArrayList<T> items = new ArrayList<T>();");
    }
    
    ValidationScanner scanner = new ValidationScanner();

    declaration.traverse(scanner, declaration.scope.classScope());
    
    if (yields.isEmpty())
    {
      invalid("No yield statements are found in the form: items.add(...);");
    }
    
    for(Scope scope: yields)
    {
      Scope yieldScope = scope;
      
      do
      {
        allScopes.put(yieldScope.node, yieldScope);
        yieldScope = yieldScope.parent;
      }
      while(yieldScope != null);
    }

    // Collect scope from breaks.
    boolean collected = !breaks.isEmpty();

    while(collected)
    {
      collected = false;
      
      for(Scope scope: breaks)
      {
        Scope target = scope.target;
        
        if (((target == null) || allScopes.containsKey(target.node)) && 
          !allScopes.containsKey(scope.node))
        {
          collected = true;
          
          Scope breakScope = scope;
          
          do
          {
            allScopes.put(breakScope.node, breakScope);
            breakScope = breakScope.parent;
          }
          while(breakScope != null);
        }
      }
    }
    
    for(Scope scope: variableDecls)
    {
      boolean stateVariable = false;
      
      if (allScopes.containsKey(scope.parent.node))
      {
        stateVariable = true;
      }
      
      if (stateVariable)
      {
        LocalDeclaration variable = (LocalDeclaration)scope.node;
        
        allScopes.put(scope.node, scope);
        stateVariables.add(fieldDef(variable.name, variable.type));
      }
    }
    
    root = allScopes.get(declaration);
    
    for(ASTNode tree: allScopes.keySet())
    {
      if (tree instanceof SynchronizedStatement)
      {
        invalid("No synchronized statement is allowed in this context.");
      }
      
      if (tree instanceof ForeachStatement)
      {
        createIterator = true;
      }
    }

    if (!stateVariables.isEmpty())
    {
      HashSet<String> names = new HashSet<String>();
      
      for(FieldDeclaration variable: stateVariables)
      {
        if (!names.add(new String(variable.name)))
        {
          invalid("No variables with duplicate names " +
            "should be used in the state machine.");
        }
      }
    }
    
    TypeReference returnMethodType = declaration.returnType;
    
    if (returnMethodType instanceof ParameterizedSingleTypeReference)
    {
      ParameterizedSingleTypeReference parameterizedType = 
        (ParameterizedSingleTypeReference)returnMethodType;
      
      itemType = parameterizedType.typeArguments[0];
    }
    else
    {
      ParameterizedQualifiedTypeReference parameterizedType = 
        (ParameterizedQualifiedTypeReference)returnMethodType;
      int length = parameterizedType.typeArguments.length;
      
      itemType = parameterizedType.typeArguments[length - 1][0];
    }
    
    statePrefix = "$state";

Prefix:
    for(int i = 1;; ++i)
    {
      for(String name: names)
      {
        if (name.startsWith(statePrefix))
        {
          statePrefix = "$state" + i;
          
          continue Prefix;
        }
      }
  
      break;
    }
  
    stateClassName = name(statePrefix);
    stateIdName = name(statePrefix + "$id");
    hasNextName = name(statePrefix + "$hasNext");
    nextDefinedName = name(statePrefix + "$nextDefined");
    nextName = name(statePrefix + "$next");
    errorName = name(statePrefix + "$exception");
    moveNextName = name(statePrefix + "$next");
    
    if (createIterator)
    {
      iteratorName = name(statePrefix + "$iterator");
    }

    stateVariables.add(fieldDef(stateIdName, type(TypeIds.T_int)));
    stateVariables.add(fieldDef(hasNextName, type(TypeIds.T_boolean)));
    stateVariables.add(fieldDef(nextDefinedName, type(TypeIds.T_boolean)));
    stateVariables.add(fieldDef(nextName, itemType));

    return true;
  }

  /**
   * Transforms method body.
   */
  private boolean refactor()
  {
    current = root; 
    
    Label begin = label();
    
    begin.constantExpression = literal(0);
    
    Label iteratorLabel = getIterationLabel(root);
    
    usedLabels.add(begin);
    usedLabels.add(iteratorLabel);
    usedLabels.add(getBreakLabel(root));

    handleRewrites();
    addStatement(begin);
    addStatement(assign(ref(stateIdName), literal(iteratorLabel)));
    addStatement(iteratorLabel);
    root.refactor();
    optimizeBreaks();
    modifyMethod();
    
    return true;
  }
  
  /**
   * Tests whether type represents ArrayList type.
   * @param type a type to test.
   * @return true if type represents ArrayList type, and 
   *   false otherwise.
   */
  private static boolean isArrayList(TypeReference type)
  {
    if ((type == null) || (type.dimensions() != 0))
    {
      return false;
    }

    if (type instanceof ParameterizedSingleTypeReference)
    {
      ParameterizedSingleTypeReference parameterizedType = 
        (ParameterizedSingleTypeReference)type;

      if (!equal(parameterizedType.token, nameArrayList))
      {
        return false;
      }
      
      TypeReference[] typeArguments = parameterizedType.typeArguments;
      
      if ((typeArguments == null) || (typeArguments.length != 1))
      {
        return false;
      }

      return true;
    }
    else if (type instanceof ParameterizedQualifiedTypeReference)
    {
      ParameterizedQualifiedTypeReference parameterizedType =
        (ParameterizedQualifiedTypeReference)type;

      if (!equal(parameterizedType.tokens, namesJavaUtilArrayList))
      {
        return false;
      }

      TypeReference[][] typeArguments = parameterizedType.typeArguments;
      
      for(int i = 0; i < typeArguments.length - 1; ++i)
      {
        if (typeArguments[i] != null)
        {
          return false;
        }
      }
      
      TypeReference[] last = typeArguments[typeArguments.length - 1];
      
      if ((last == null) || (last.length != 1))
      {
        return false;
      }
      
      return true;
    }
    else
    {
      return false;
    }
  }

  /**
   * Tests whether expression is of the form: ArrayList<T>().
   * @param expression an expression to test.
   * @return true if exression of the form ArrayList<T>(), and 
   *   false otherwise.
   */
  private static boolean isNewArrayList(Expression expression)
  {
    if (expression instanceof AllocationExpression)
    {
      AllocationExpression allocation = (AllocationExpression)expression;
      
      if (!(allocation instanceof QualifiedAllocationExpression) &&
        (allocation.arguments == null) &&
        (allocation.typeArguments == null) &&
        isArrayList(allocation.type))
      {
        return true;
      }
    }
    
    return false;
  }
  
  /**
   * Validates return statement.
   * @param expression an expression to test.
   * @return true if it's a valid return statement, and false otherwise.
   */
  private boolean validateReturn(ReturnStatement statement)
  {
    Expression expression = statement.expression;
    
    if (iterable)
    {
      if (expression instanceof SingleNameReference)
      {
        SingleNameReference reference = (SingleNameReference)expression;
        
        if (equal(reference.token, iteratorDecl.name))
        {
          return true;
        }
      }
    }
    else
    {
      if (expression instanceof MessageSend)
      {
        MessageSend invoke = (MessageSend)expression;
        
        if (equal(invoke.selector, nameIterator))
        {
          if (invoke.receiver instanceof SingleNameReference)
          {
            SingleNameReference reference = 
              (SingleNameReference)invoke.receiver;
            
            if (equal(reference.token, iteratorDecl.name))
            {
              return true;
            }
          }
        }
      }
    }
    
    return false;
  }
  
  /**
   * Gets yield expression from yield statement: iterator.add(...);
   * @param invoke an invoke expression.
   * @return a yield expression, or null if it's not a yield statement.
   */
  private Expression getYieldExpression(MessageSend invoke)
  {
    if ((invoke.statementEnd != -1) && 
      (invoke.typeArguments == null) &&
      equal(invoke.selector, nameAdd) &&
      (invoke.arguments != null) && 
      (invoke.arguments.length == 1) &&
      (invoke.receiver instanceof SingleNameReference))
    {
      SingleNameReference reference = 
        (SingleNameReference)invoke.receiver;
      
      if (equal(reference.token, iteratorDecl.name))
      {
        return invoke.arguments[0];
      }
    }
    
    return null;
  }
  
  /**
   * Gets set stated id, if any.
   * @param statement a statement to test.
   * @return label expression, or null if statement 
   *   is not of the specified form.
   */
  private Expression getSetStateId(Statement statement)
  {
    if (statement instanceof Assignment)
    {
      Assignment assign = (Assignment)statement;
      
      if (assign.lhs instanceof SingleNameReference)
      {
        SingleNameReference field = (SingleNameReference)assign.lhs;
        
        // Note reference comparison here.
        if (field.token == stateIdName)
        {
          return assign.expression;
        }
      }
    }
    
    return null;
  }
  
  /**
   * Tests whether the statement is continue;
   * @param statement a statement to test.
   * @return true if statement is continue, and false otherwise.
   */
  private static boolean isContinue(Statement statement)
  {
    if (statement instanceof ContinueStatement)
    {
      ContinueStatement continueStatement = (ContinueStatement)statement;
      
      if (continueStatement.label == null)
      {
        return true;
      }
    }
    
    return false;
  }
  
  /**
   * Create a continue label.
   * @return a continue label.
   */
  private ContinueStatement doContinue()
  {
    return source(new ContinueStatement(null, 0, 0));    
  }
  
  /**
   * Gets a label from label expression.
   * @param expression a label expression.
   * @return a label or null if expression is not literal label.
   */
  private Label getLabel(Expression expression)
  {
    return labelLiterals.get(expression);
  }
  
  /**
   * Creates a new label.
   * @return a new label.
   */
  private Label label()
  {
    return new Label();
  }
  
  /**
   * Adds a label statement.
   * @param label a label statement.
   */
  private void addStatement(Label label)
  {
    label.id = statements.size();
    statements.add(source(label));
  }
  
  /**
   * Adds a statement to a current case.
   * @param statement a case statement.
   */
  private void addStatement(Statement statement)
  {
    statements.add(source(statement));
  }

  /**
   * Gets values as array of arrays of chars.
   * @param values a name value.
   * @return an array of arrays of chars.
   */
  private static char[][] names(String... values)
  {
    if (values == null)
    {
      return null;
    }
    
    char[][] names = new char[values.length][];
    
    for(int i = 0; i < values.length; ++i)
    {
      names[i] = name(values[i]);
    }
    
    return names;
  }
  
  /**
   * Gets value as array of chars.
   * @param value a name value.
   * @return an array of chars.
   */
  private static char[] name(String value)
  {
    if (value == null)
    {
      return null;
    }
    
    return value.toCharArray();
  }

  /**
   * Gets an array of positions for names.
   * @param names an array of names.
   * @return an array of positions.
   */
  private static long[] positions(char[][] names)
  {
    return names == null ? null : new long[names.length];
  }
  
  /**
   * Tests whether two names are equal.
   * @param first a first name.
   * @param second a second name.
   * @return true if names are equal, and false otherwise.
   */
  private static boolean equal(char[] first, char[] second)
  {
    return Arrays.equals(first, second);
  }

  /**
   * Tests whether two names are equal.
   * @param first a first name.
   * @param second a second name.
   * @return true if names are equal, and false otherwise.
   */
  private static boolean equal(char[][] first, char[][] second)
  {
    if (first == second)
    {
      return true;
    }
    
    if ((first == null) || (second == null))
    {
      return false;
    }
    
    int length = first.length;

    if (second.length != length)
    {
      return false;
    }

    for(int i = 0; i < length; ++i)
    {
      if (!equal(first[i], second[i]))
      {
        return false;
      }
    }

    return true;
  }
  
  /**
   * Assigns a source start and end positions.
   * @param <T> a node type.
   * @param node a node instance.
   * @param reference a reference node.
   * @return a node instance.
   */
  private <T extends ASTNode> T source(T node)
  {
    if ((node != null) && (current !=null) && (node.sourceStart == 0))
    {
      node.sourceStart = current.node.sourceStart;
      node.sourceEnd = current.node.sourceEnd;
      
      if (node instanceof AbstractMethodDeclaration)
      {
        AbstractMethodDeclaration declaration =
          (AbstractMethodDeclaration)node;
        
        if (current.node instanceof AbstractMethodDeclaration)
        {
          AbstractMethodDeclaration currentDeclaration = 
            (AbstractMethodDeclaration)current.node;
          
          declaration.bodyStart = currentDeclaration.bodyStart;
          declaration.bodyEnd = currentDeclaration.bodyEnd;
          declaration.declarationSourceStart = 
            currentDeclaration.declarationSourceStart;
          declaration.declarationSourceEnd = 
            currentDeclaration.declarationSourceEnd;
        }
      }
    }
    
    return node;
  }
  
  /**
   * Creates a {@link TypeReference} instance for a type id.
   * @param typeId a type id.
   * @return a {@link TypeReference} instance.
   */
  private TypeReference type(int typeId)
  {
    return TypeReference.baseTypeReference(typeId, 0);
  }

  /**
   * Creates a {@link TypeReference} instance.
   * @param name a type name.
   * @param arguments optional type arguments.
   * @return a {@link TypeReference} instance.
   */
  private TypeReference type(char[] name, TypeReference ...arguments)
  {
    if ((arguments == null) || (arguments.length == 0))
    {
      return new SingleTypeReference(name, 0); 
    }
    else
    {
      return new ParameterizedSingleTypeReference(name, arguments, 0, 0);
    }
  }

  /**
   * Creates a {@link TypeReference} instance.
   * @param name a type name.
   * @param arguments optional type arguments.
   * @return a {@link TypeReference} instance.
   */
  private TypeReference type(char[][] name, TypeReference... arguments)
  {
    long[] pos = positions(name);
    
    if ((arguments == null) || (arguments.length == 0))
    {
      return new QualifiedTypeReference(name, pos); 
    }
    else
    {
      TypeReference[][] types = new TypeReference[name.length][];
      
      types[types.length - 1] = arguments;
      
      return new ParameterizedQualifiedTypeReference(name, types, 0, pos);
    }
  }
  
  /**
   * Gets a variable reference.
   * @param variable a variable name.
   * @return a {@link SingleNameReference}.
   */
  private Reference ref(char[] variable)
  {
    return source(new SingleNameReference(variable, 0));
  }

  /**
   * Creates an assignment statement.
   * @param left a left side assignmet.
   * @param right a right side assignment.
   * @return an assignment statement.
   */
  private Statement assign(Expression left, Expression right)
  {
    return source(new Assignment(source(left), source(right), 0));
  }

  /**
   * Creates a block from statement.
   * @param statements a list of statement.
   * @return a block instance.
   */
  private Block block(Statement... statements)
  {
    Block block = new Block(0);
    
    if (statements != null && statements.length > 0)
    {
      for(Statement statement: statements)
      {
        source(statement);
      }

      block.statements = statements;
    }
    
    return source(block);
  }
  
  /**
   * Creates an array of statements.
   * @param statements a list of statement.
   * @return an array of statements.
   */
  private Statement[] statements(Statement... statements)
  {
    if (statements != null)
    {
      for(Statement statement: statements)
      {
        source(statement);
      }
    }

    return statements;
  }

  /**
   * Creates a return statement.
   * @param expression a return value.
   * @return a return statement.
   */
  private Statement ret(Expression expression)
  {
    return source(new ReturnStatement(source(expression), 0, 0));
  }

  /**
   * Creates if-then statement.
   * @param condition a condition expression.
   * @param statement a then statement.
   * @return {@link JCIf} statement.
   */
  private Statement ifThen(Expression condition, Statement then)
  {
    return source(new IfStatement(source(condition), source(then), 0, 0));
  }

  /**
   * Creates new Class() expression.
   * @param type a class type.
   * @param arguments arguments.
   * @return a new Class() expression.
   */
  private AllocationExpression newClass(
    TypeReference type, 
    Expression... arguments)
  {
    AllocationExpression allocation = new AllocationExpression();
    
    allocation.type = type;
    
    if ((arguments != null) && (arguments.length > 0))
    {
      for(Expression argument: arguments)
      {
        source(argument);
      }
      
      allocation.arguments = arguments;
    }
    
    return source(allocation);
  }

  /**
   * Creates a method definition.
   * @param modifiers method modifiers.
   * @param name method name.
   * @param statements method statements.
   * @param returnType a return type.
   * @param params method parameters.
   * @return a method definition.
   */
  private MethodDeclaration methodDef(
    int modifiers,
    char[] name,
    Statement[] statements,
    TypeReference returnType,
    Argument... arguments)
  {
    MethodDeclaration method = 
      new MethodDeclaration(declaration.compilationResult);
    
    if (statements != null)
    {
      for(Statement statement: statements)
      {
        source(statement);
      }
    }

    method.selector = name;
    method.modifiers = modifiers;
    method.arguments = arguments;
    method.returnType = returnType;
    method.statements = statements;
    
    return source(method);
  }
  
  /**
   * Creates a field definition.
   * @param name a variable name.
   * @param type a variable type.
   * @return a field definition.
   */
  private FieldDeclaration fieldDef(char[] name, TypeReference type)
  {
    FieldDeclaration field = new FieldDeclaration(name, 0, 0);
    
    field.type = type;
    field.modifiers = ClassFileConstants.AccPrivate;
    
    return field;
  }

  /**
   * Creates a variable definition.
   * @param name a variable name.
   * @param type a variable type.
   * @return a variable definition.
   */
  private LocalDeclaration varDef(char[] name, TypeReference type)
  {
    LocalDeclaration local = new LocalDeclaration(name, 0, 0);
    
    local.type = type;
    
    return source(local);
  }
  
  /**
   * Creates an integer literal.
   * @param value a literal value.
   * @return an integer literal.
   */
  private IntLiteral literal(int value)
  {
    return source(IntLiteral.buildIntLiteral(
      String.valueOf(value).toCharArray(), 0, 0));
  }

  /**
   * Creates a boolean literal.
   * @param value a literal value.
   * @return a boolean literal.
   */
  private Literal literal(boolean value)
  {
    return source(value ? new TrueLiteral(0, 0) : new FalseLiteral(0, 0));
  }

  /**
   * Creates a null literal.
   * @return a null literal.
   */
  private Literal literal()
  {
    return source(new NullLiteral(0, 0));
  }

  /**
   * Creates a label literal. 
   * @param label a label reference.
   */
  private Expression literal(Label label)
  {
    Expression literal = new UnaryExpression(
      label.constantExpression != null ? label.constantExpression : MinusOne, 
      OperatorIds.PLUS);
    
    labelLiterals.put(literal, label);
    
    return source(literal);
  }
  
  /**
   * Gets a break label.
   * @param scope a scope to get break label for.
   * @return a break label.
   */
  public Label getBreakLabel(Scope scope)
  {
    Label label = scope.breakLabel;
    
    if (label == null)
    {
      label = label();
      scope.breakLabel = label;
    }
    
    return label;
  }

  /**
   * Gets a break label.
   * @param scope a scope to get finally label for.
   * @return a break label.
   */
  public Label getFinallyLabel(Scope scope)
  {
    Label label = scope.finallyLabel;
    
    if (label == null)
    {
      label = label();
      scope.finallyLabel = label;
    }
    
    return label;
  }

  /**
   * Gets an iteration label.
   * @param scope a scope to get iteration label for.
   * @return an iteration label.
   */
  public Label getIterationLabel(Scope scope)
  {
    Label label = scope.iterationLabel;
    
    if (label == null)
    {
      label = label();
      scope.iterationLabel = label;
    }
    
    return label;
  }
  
  /**
   * Gets closest try scope with finally clause, if any.
   * @param scope a start scope.
   * @param top optional top scope.
   * @return closest try scope with finally clause, if any.
   */
  private static Scope getFinallyScope(Scope scope, Scope top)
  {
    ASTNode previous = null;
    
    while(scope != null)
    {
      ASTNode tree = scope.node;
      
      if (tree instanceof TryStatement)
      {
        TryStatement statement = (TryStatement)tree;
        
        if ((statement.finallyBlock != null) && 
          (statement.finallyBlock != previous))
        {
          return scope;
        }
      }
      
      if (scope == top)
      {
        break;
      }
      
      previous = tree;
      scope = scope.parent;
    }
    
    return null;
  }
  
  /**
   * Invokes a method.
   * @param instance a method instance.
   * @param name a method name.
   * @param arguments method arguments.
   * @return an invoke expression.
   */
  private Expression invoke(
    Expression instance, 
    char[] name, 
    Expression... arguments)
  {
    MessageSend invoke = new MessageSend();
    
    invoke.receiver = 
      source(instance == null ? ThisReference.implicitThis() : instance);
    invoke.selector = name;
    
    if (arguments != null && arguments.length > 0)
    {
      for(Expression argument: arguments)
      {
        source(argument);
      }

      invoke.arguments = arguments;
    }
    
    return source(invoke);
  }

  
  /**
   * Creates an annotation to suppress all warnings.
   * @return An annotation to suppress all warnings.
   */
  private Annotation suppressWarnings()
  {
    SingleMemberAnnotation suppressWarnings = 
      new SingleMemberAnnotation(type(namesJavaLangSuppressWarnings), 0);
    
    suppressWarnings.memberValue = new StringLiteral(nameAll, 0, 0, 0);
    
    return suppressWarnings;
  }
  /**
   * Refactors a statement.
   * @param statement a statement to refactor.
   */
  private void refactorStatement(Statement statement)
  {
    if (statement == null)
    {
      return;
    }

    Scope scope = allScopes.get(statement);
    
    if (scope != null)
    {
      Scope previous = current;
      
      current = scope;
      scope.refactor();
      current = previous;
    }
    else
    {
      if (statement != iteratorDecl)
      {
        addStatement(statement);        
      }
    }
  }
  
  /**
   * Rewrites code.
   */
  private void handleRewrites()
  {
    for(Scope scope: rewrites)
    {
      scope.refactor();
    }
  }

  /**
   * Optimizes break statements.
   */
  private void optimizeBreaks()
  {
    int count = statements.size();
    
    for(Map.Entry<Expression, Label> entry: labelLiterals.entrySet())
    {
      Label label = entry.getValue();
      int i = label.id;
        
      while((i >= 0) && (i + 1 < count))
      {
        Statement statement = statements.get(++i);
        
        if (statement instanceof Label)
        {
          label = (Label)statement;
          
          continue;
        }
        
        Label next = getLabel(getSetStateId(statement));
        
        if (next != null)
        {
          statement = i + 1 < count ? statements.get(i + 1) : null;
          
          if ((statement == null) || 
            isContinue(statement) || 
            (statement instanceof Label))
          {
            label = next;
            i = label.id;
            
            continue;
          }
        }
        
        break;
      }
      
      entry.setValue(label);
      
      if (label.id >= 0)
      {
        usedLabels.add(label);
      }
    }

    int id = 0;
    Statement last = null;
    
    for(int i = 0; i < count; ++i)
    {
      Statement statement = statements.get(i);
      
      if (statement == null)
      {
        continue;
      }
      else if (statement instanceof Label)
      {
        Label label = (Label)statement;
        
        if (!usedLabels.contains(label))
        {
          statements.set(i, null);
          
          if ((isContinue(last) || (last instanceof ReturnStatement)) &&
            (i + 1 < count) && 
            getSetStateId(statements.get(i + 1)) != null)
          {
            statements.set(i + 1, null);
            
            if ((i + 2 < count) && isContinue(statements.get(i + 2)))
            {
              statements.set(i + 2, null);
            }
          }
        }
        else
        {
          label.id = id++;
          label.constantExpression = literal(label.id);
        }
      }
      else
      {  
        if ((last != null) && isContinue(statement))
        {
          Label previous = getLabel(getSetStateId(last));
          
          if (previous != null)
          {
            for(int j = i; ++j < count;)
            {
              statement = statements.get(j);
              
              if (statement == null)
              {
                continue;
              }
              
              if (statement instanceof Label)
              {
                Label next = (Label)statement;
                
                if (next == previous)
                {
                  statement = last;
                  statements.set(i, null);
                }
              }
  
              break;
            }
          }
        }
        
        last = statement;
      }
    }
    
    // Adjust label literals
    for(Map.Entry<Expression, Label> entry: labelLiterals.entrySet())
    {
      UnaryExpression expression = (UnaryExpression)entry.getKey();
      Label label = entry.getValue();
      
      expression.expression = literal(label.id);
    }
  }
  
  /**
   * Modifies original method.
   */
  private void modifyMethod()
  {
    declaration.statements = statements(
      generateTypeAssert(),
      generateIteratorClass(),
      ret(newClass(type(stateClassName))));
  }
  
  /**
   * Generates verification assert.
   * @return a verification assert.
   */
  private Statement generateTypeAssert()
  {
    return source(
      new AssertStatement(
        new EqualExpression(
          new CastExpression(
            new CastExpression(literal(), iteratorDecl.type), 
            type(namesJavaUtilArrayList, itemType)), 
          literal(), 
          OperatorIds.EQUAL_EQUAL), 
        0));
  }

  /**
   * Generates iterator instance.
   * @return iterator instance.
   */
  private TypeDeclaration generateIteratorClass()
  {
    TypeDeclaration type = new TypeDeclaration(declaration.compilationResult);
    
    type.name = stateClassName;
    
    type.superInterfaces = new TypeReference[]
    {
      type(namesJavaLangIterable, itemType),
      type(namesJavaUtilIterator, itemType),
      type(namesJavaIoCloseable)
    };
    
    if (!classes.isEmpty())
    {
      type.memberTypes = classes.toArray(new TypeDeclaration[classes.size()]);
    }
    
    AbstractMethodDeclaration[] methods = 
      new AbstractMethodDeclaration[createIterator ? 10 : 7];
    
    methods[0] = generateStateConstructor();
    methods[1] = generateIterator();
    methods[2] = generateHasNext();
    methods[3] = generateNext();
    methods[4] = generateRemove();
    methods[5] = generateClose();
    methods[6] = generateMoveNext();
      
    if (createIterator)
    {
      methods[7] = generateIteratorIterator();
      methods[8] = generateObjectArrayIterator();
      methods[9] = generateArrayIterator();
    }
    
    type.methods = methods;
    type.fields = 
      stateVariables.toArray(new FieldDeclaration[stateVariables.size()]);

    return source(type);
  }

  /**
   * Generates state constructor.
   * @return state constructor.
   */
  private ConstructorDeclaration generateStateConstructor()
  {
    ConstructorDeclaration constructor = 
      new ConstructorDeclaration(declaration.compilationResult);
    
    constructor.selector = stateClassName;
    constructor.constructorCall = 
      new ExplicitConstructorCall(ExplicitConstructorCall.ImplicitSuper);
    
    return source(constructor);    
  }

  /**
   * Generates iterator() method.
   * @return iterator() method.
   */
  private MethodDeclaration generateIterator()
  {
    return methodDef(
      ClassFileConstants.AccPublic, 
      nameIterator,
      statements(
        new IfStatement(
          new EqualExpression(
            ref(stateIdName), 
            literal(0), 
            OperatorIds.EQUAL_EQUAL), 
          block(
            assign(ref(stateIdName), literal(1)), 
            ret(new ThisReference(0, 0))), 
          ret(newClass(type(stateClassName))), 
          0, 
          0)),
      type(namesJavaUtilIterator, itemType));    
  }
  
  /**
   * Generates moveNext() method.
   * @return moveNext() method.
   */
  private MethodDeclaration generateMoveNext()
  {
    SwitchStatement switchStatement = generateSwitch();
    Statement errorHandler = generateErrorHandler();
    Statement[] statements;
    
    if (errorHandler != null)
    {
      TryStatement tryStatement = new TryStatement();
      
      tryStatement.tryBlock = block(switchStatement);
      
      tryStatement.catchArguments = new Argument[]
      {
        new Argument(nameE, 0, type(namesJavaLangThrowable), 0)
      };
      
      tryStatement.catchBlocks = new Block[]
      {
        block(assign(ref(errorName), ref(nameE)))
      };
      
      statements = statements(
        varDef(errorName, type(namesJavaLangThrowable)),
        new WhileStatement(
          literal(true), 
          block(tryStatement, errorHandler), 
          0, 
          0));
    }
    else
    {
      statements = statements(
        new WhileStatement(literal(true), switchStatement, 0, 0));
    }

    return methodDef(
      ClassFileConstants.AccPrivate,
      moveNextName,
      statements,
      type(TypeIds.T_boolean));
  }


  /**
   * Generates iterator() method for an iterable.
   * @return iterator() method.
   */
  private MethodDeclaration generateIteratorIterator()
  {
    MethodDeclaration method = methodDef(
      ClassFileConstants.AccPrivate,
      iteratorName,
      statements(ret(invoke(ref(nameValue), nameIterator))),
      type(namesJavaUtilIterator),
      new Argument(nameValue, 0, type(namesJavaLangIterable), 0));
    
    method.annotations = new Annotation[] { suppressWarnings() };

    return method;
  }

  /**
   * Generates iterator() method for an Object[].
   * @return iterator() method.
   */
  private MethodDeclaration generateObjectArrayIterator()
  {
    MethodDeclaration method = methodDef(
      ClassFileConstants.AccPrivate,
      iteratorName,
      statements(
        ret(
          invoke(
            invoke(
              new QualifiedNameReference(
                namesJavaUtilArrays, 
                positions(namesJavaUtilArrays), 
                0, 
                0),
                nameAsList,
                ref(nameValue)),
            nameIterator))),
      type(namesJavaUtilIterator),
      new Argument(
        nameValue, 
        0, 
        new ArrayQualifiedTypeReference(
          namesJavaLangObject, 
          1, 
          positions(namesJavaLangObject)),
        0));
    
    method.annotations = new Annotation[] { suppressWarnings() };

    return method;
  }

  /**
   * Generates iterator() method for an array.
   * @return iterator() method.
   */
  private MethodDeclaration generateArrayIterator()
  {
    MethodDeclaration hasNextMethod = 
      methodDef(
        ClassFileConstants.AccPublic, 
        nameHasNext, 
        statements(
          ret(
            new BinaryExpression(
              ref(nameIndex), 
              ref(nameLength), 
              OperatorIds.LESS))), 
        type(TypeIds.T_boolean));  
      
    MethodDeclaration nextMethod = 
      methodDef(
        ClassFileConstants.AccPublic, 
        nameNext, 
        statements(
          ret(
            invoke(
              new QualifiedNameReference(
                namesJavaLangReflectArray, 
                positions(namesJavaLangReflectArray), 
                0, 
                0),
              nameGet,
              ref(nameValue),
              new PostfixExpression(
                ref(nameIndex), 
                IntLiteral.One, 
                OperatorIds.PLUS, 
                0)))), 
        type(namesJavaLangObject));  
    
    MethodDeclaration removeMethod = generateRemove(); 

    FieldDeclaration indexVar = fieldDef(nameIndex, type(TypeIds.T_int));
    LocalDeclaration lengthVar = new LocalDeclaration(nameLength, 0, 0);
    
    lengthVar.modifiers = ClassFileConstants.AccFinal;
    lengthVar.type = type(TypeIds.T_int);
    lengthVar.initialization =
      invoke(
        new QualifiedNameReference(
          namesJavaLangReflectArray, 
          positions(namesJavaLangReflectArray), 
          0, 
          0),
        nameGetLength, 
        ref(nameValue));
    
    TypeDeclaration type = new TypeDeclaration(declaration.compilationResult);
    
    TypeReference base = type(namesJavaUtilIterator);

    type.name = new char[0];
    type.superInterfaces = new TypeReference[] { base };
    type.fields = new FieldDeclaration[] { indexVar };
    
    type.methods = new AbstractMethodDeclaration[]
    {
      hasNextMethod,
      nextMethod,
      removeMethod
    };
    
    type.bits |= ASTNode.IsAnonymousType | ASTNode.IsLocalType;

    QualifiedAllocationExpression allocation = 
      new QualifiedAllocationExpression(type);
    
    allocation.type = base;
    
    MethodDeclaration method = methodDef(
      ClassFileConstants.AccPrivate,
      iteratorName,
      statements(
        lengthVar,
        ret(allocation)),
      type(namesJavaUtilIterator),
      new Argument(
        nameValue, 
        0, 
        type(namesJavaLangObject), 
        ClassFileConstants.AccFinal));
    
    method.annotations = new Annotation[] { suppressWarnings() };

    return method;
  }

  /**
   * Generates hasNext() method.
   * @return hasNext() method.
   */
  private MethodDeclaration generateHasNext()
  {
    return methodDef(
      ClassFileConstants.AccPublic,
      nameHasNext,
      statements(
        ifThen(
          new UnaryExpression(ref(nextDefinedName), OperatorIds.NOT), 
          block(
            assign(ref(hasNextName), invoke(null, moveNextName)),
            assign(ref(nextDefinedName), literal(true)))),
        ret(ref(hasNextName))),
      type(TypeIds.T_boolean));
  }

  /**
   * Generates next() method.
   * @return next() method.
   */
  private MethodDeclaration generateNext()
  {
    return methodDef(
      ClassFileConstants.AccPublic,
      nameNext,
      statements(
        ifThen(
          new UnaryExpression(invoke(null, nameHasNext), OperatorIds.NOT),
          new ThrowStatement(
            newClass(type(namesJavaUtilNoSuchElementException)), 0, 0)),
        assign(ref(nextDefinedName), literal(false)),
        ret(ref(nextName))),
      itemType);
  }

  /**
   * Generates remove() method.
   * @return remove() method.
   */
  private MethodDeclaration generateRemove()
  {
    return methodDef(
      ClassFileConstants.AccPublic,
      nameRemove,
      statements(
        new ThrowStatement(
          newClass(type(namesJavaLangUnsupportedOperationException)),
          0,
          0)),
      type(TypeIds.T_void));
  }

  /**
   * Generates close() method.
   * @return close() method.
   */
  private MethodDeclaration generateClose()
  {
    Statement statement;
    
    if (breakCases.isEmpty())
    {
      statement = assign(ref(stateIdName), literal(getBreakLabel(root)));
    }
    else
    {
      Label prevLabel = null;
      int prevIndex = -1;
      
      for(int i = 0; i < breakCases.size(); ++i)
      {
        statement = breakCases.get(i);
        
        if (statement instanceof CaseStatement)
        {
          CaseStatement caseStatement = (CaseStatement)statement;
          Label label = getLabel(caseStatement.constantExpression);
          
          if ((prevLabel != null) && (prevLabel == label))
          {
            for(int j = i; j-- > prevIndex;)
            {
              breakCases.remove(j);
            }
            
            i = prevIndex;
          }
          else
          {
            prevLabel = label;
            prevIndex = i;
          }
        }
      }
      
      breakCases.add(new CaseStatement(null, 0, 0));
      breakCases.add(assign(ref(stateIdName), literal(getBreakLabel(root))));
      breakCases.add(new ReturnStatement(null, 0, 0));
      
      SwitchStatement switchStatement = new SwitchStatement();
      
      switchStatement.expression = ref(stateIdName);
      switchStatement.statements = 
        breakCases.toArray(new Statement[breakCases.size()]);
      
      statement = new DoStatement(
        invoke(null, moveNextName), 
        switchStatement, 
        0, 
        0);
    }
    
    return methodDef(
      ClassFileConstants.AccPublic,
      nameClose,
      statements(statement),
      type(TypeIds.T_void));
  }

  /**
   * Generates switch statements.
   * @return a switch statement.
   */
  private SwitchStatement generateSwitch()
  {
    int count = 2;
    
    for(Statement statement: statements)
    {
      if (statement != null)
      {
        ++count;
      }
    }
    
    Statement[] cases = new Statement[count];
    int i = 0;
    
    for(Statement statement: statements)
    {
      if (statement != null)
      {
        cases[i++] = statement;
      }
    }
    
    cases[i++] = new CaseStatement(null, 0, 0);
    cases[i++] = ret(literal(false));
    
    SwitchStatement switchStatement = new SwitchStatement();
    
    switchStatement.expression = ref(stateIdName);
    switchStatement.statements = cases;
    
    return source(switchStatement);
  }
  
  /**
   * Generates a switch for the error handlers.
   * @return a switch for the error handlers.
   */
  private Statement generateErrorHandler()
  {
    if (errorHandlers.isEmpty())
    {
      return null;
    }
    
    ArrayList<Statement> cases = new ArrayList<Statement>();
    HashSet<Label> labels = new HashSet<Label>();
    
    for(ErrorHandler handler: errorHandlers)
    {
      for(int i = handler.begin; i < handler.end; ++i)
      {
        Statement statement = statements.get(i);
        
        if (statement instanceof Label)
        {
          Label label = (Label)statement;
          
          if (labels.add(label))
          {
            cases.add(label);
          }
        }
      }
      
      cases.addAll(handler.statements);
    }
    
    LocalDeclaration exception = new LocalDeclaration(nameE, 0, 0);
    
    exception.type = type(namesJavaUtilConcurrentModificationException);
    exception.initialization = newClass(exception.type);
    
    cases.add(new CaseStatement(null, 0, 0));
    cases.add(assign(ref(stateIdName), literal(getBreakLabel(root))));
    cases.add(exception);
    cases.add(invoke(ref(exception.name), nameInitCause, ref(errorName)));
    cases.add(new ThrowStatement(ref(exception.name), 0, 0));
    
    SwitchStatement switchStatement = new SwitchStatement();
    
    switchStatement.expression = ref(stateIdName);
    switchStatement.statements = cases.toArray(new Statement[cases.size()]);

    return source(switchStatement);
  }
  
  /**
   * A processing environment.
   */
  private final BaseProcessingEnvImpl environment;
  
  /**
   * A method element.
   */
  private final ExecutableElement method;
  
  /**
   * A method declaration. 
   */
  private final MethodDeclaration declaration;
  
  /**
   * Type utils.
   */
  private final Types types;
  
  /**
   * Element utils.
   */
  private final Elements elements;

  /**
   * Item type.
   */
  private TypeReference itemType;
  
  /**
   * True if method is Iterable<T>, and false if Iterator<T>.
   */
  private boolean iterable;
  
  /**
   * Iterator declaration.
   */
  private LocalDeclaration iteratorDecl;
  
  /**
   * All names used.
   */
  private HashSet<String> names = new HashSet<String>();
  
  /**
   * Local classes.
   */
  private ArrayList<TypeDeclaration> classes = new ArrayList<TypeDeclaration>();

  /**
   * A list of yield scopes.
   */
  private ArrayList<Scope> yields = new ArrayList<Scope>();
  
  /**
   * Break and continue statements.
   */
  private ArrayList<Scope> breaks = new ArrayList<Scope>();
  
  /**
   * Rewrite scopes.
   */
  private ArrayList<Scope> rewrites = new ArrayList<Scope>();

  /**
   * Variable declarations.
   */
  private ArrayList<Scope> variableDecls = new ArrayList<Scope>();
  
  /**
   * Variable declarations.
   */
  private ArrayList<FieldDeclaration> stateVariables = 
    new ArrayList<FieldDeclaration>();

  /**
   * A root scope.
   */
  private Scope root;

  /**
   * Current scope.
   */
  private Scope current;

  /**
   * All scopes.
   */
  private HashMap<ASTNode, Scope> allScopes = new HashMap<ASTNode, Scope>();

  /**
   * Refactored method body.
   */
  private ArrayList<Statement> statements = new ArrayList<Statement>();
  
  /**
   * Refactored method body.
   */
  private ArrayList<Statement> breakCases = new ArrayList<Statement>();
  
  /**
   * Refactored method exception handlers.
   */
  private ArrayList<ErrorHandler> errorHandlers = new ArrayList<ErrorHandler>();
  
  /**
   * Finally exceptions. 
   */
  private int finallyBlocks;
  
  /**
   * Indicates that iterator method should be created.
   */
  private boolean createIterator;

  /**
   * Map of literals to labels.
   */
  private HashMap<Expression, Label> labelLiterals = 
    new HashMap<Expression, Label>();

  /**
   * Handler labels.
   */
  private HashSet<Label> usedLabels = new HashSet<Label>();

  /**
   * A state field prefix.
   */
  private String statePrefix;
  
  /**
   * State class name.
   */
  private char[] stateClassName;
  
  /**
   * State id name.
   */
  private char[] stateIdName;

  /**
   * hasNext name.
   */
  private char[] hasNextName;

  /**
   * nextDefined name.
   */
  private char[] nextDefinedName;

  /**
   * Next value name.
   */
  private char[] nextName;

  /**
   * Error name.
   */
  private char[] errorName;

  /**
   * moveNext name.
   */
  private char[] moveNextName;

  /**
   * iterator method name.
   */
  private char[] iteratorName;

  /**
   * A -1 literal.
   */
  private final IntLiteral MinusOne = literal(-1);

  /**
   * An "iterator" name.
   */
  private static final char[] nameIterator = name("iterator");

  /**
   * An "add" name.
   */
  private static final char[] nameAdd = name("add");

  /**
   * An "e" name.
   */
  private static final char[] nameE = name("e");

  /**
   * An "value" name.
   */
  private static final char[] nameValue = name("value");

  /**
   * An "index" name.
   */
  private static final char[] nameIndex = name("index");

  /**
   * An "length" name.
   */
  private static final char[] nameLength = name("length");

  /**
   * An "getLength" name.
   */
  private static final char[] nameGetLength = name("getLength");

  /**
   * An "get" name.
   */
  private static final char[] nameGet = name("get");

  /**
   * An "close" name.
   */
  private static final char[] nameClose = name("close");

  /**
   * An "hasNext" name.
   */
  private static final char[] nameHasNext= name("hasNext");

  /**
   * An "next" name.
   */
  private static final char[] nameNext= name("next");

  /**
   * An "remove" name.
   */
  private static final char[] nameRemove = name("remove");

  /**
   * An "initCause" name.
   */
  private static final char[] nameInitCause = name("initCause");

  /**
   * An "ArrayList" names.
   */
  private static final char[] nameArrayList = name("ArrayList");
  
  /**
   * An "asList" names.
   */
  private static final char[] nameAsList = name("asList");

  /**
   * An "all" names.
   */
  private static final char[] nameAll = name("all");
  
  /**
   * An "java.io.Object" names.
   */
  private static final char[][] namesJavaLangObject =
    names("java", "lang", "Object");
  
  /**
   * An "java.lang.reflect.Array" names.
   */
  private static final char[][] namesJavaLangReflectArray =
    names("java", "lang", "reflect", "Array");

  /**
   * An "java.util.ArrayList" names.
   */
  private static final char[][] namesJavaUtilArrayList = 
    names("java", "util", "ArrayList");

  /**
   * An "java.util.NoSuchElementException" names.
   */
  private static final char[][] namesJavaUtilNoSuchElementException = 
    names("java", "util", "NoSuchElementException");

  /**
   * An "java.lang.Iterable" names.
   */
  private static final char[][] namesJavaLangIterable =
    names("java", "lang", "Iterable");
  
  /**
   * An "java.util.Iterator" names.
   */
  private static final char[][] namesJavaUtilIterator =
    names("java", "util", "Iterator");

  /**
   * An "java.io.Closeable" names.
   */
  private static final char[][] namesJavaIoCloseable =
    names("java", "io", "Closeable");
  
  /**
   * An "java.util.ConcurrentModificationException" names.
   */
  private static final char[][] namesJavaUtilConcurrentModificationException =
    names("java", "util", "ConcurrentModificationException");
  
  /**
   * An "java.lang.UnsupportedOperationException" names.
   */
  private static final char[][] namesJavaLangUnsupportedOperationException =
    names("java", "lang", "UnsupportedOperationException");

  /**
   * An "java.lang.Throwable" names.
   */
  private static final char[][] namesJavaLangThrowable =
    names("java", "lang", "Throwable");
  
  /**
   * An "java.util.Arrays" names.
   */
  private static final char[][] namesJavaUtilArrays =
    names("java", "util", "Arrays");

  /**
   * An "java.lang.SuppressWarnings" names.
   */
  private static final char[][] namesJavaLangSuppressWarnings =
    names("java", "lang", "SuppressWarnings");
  
  /**
   * A source tree validator.
   */
  private class ValidationScanner extends ASTVisitor
  {
    @Override
    public boolean visit(
      final MethodDeclaration methodDeclaration, 
      final ClassScope scope)
    {
      current = new Scope(current, methodDeclaration)
      {
        @Override
        public void refactor()
        {
          if (methodDeclaration.statements != null)
          {
            for(Statement statement: methodDeclaration.statements)
            {
              refactorStatement(statement);
            }
          }
          
          addStatement(getBreakLabel(this));
        }
      };
      
      return super.visit(methodDeclaration, scope);
    }
    
    @Override
    public void endVisit(MethodDeclaration methodDeclaration, ClassScope scope) 
    {
      current = current.parent;
    }
    
    @Override
    public boolean visit(final Block block, final BlockScope scope)
    {
      current = new Scope(current, block)
      {
        @Override
        public void refactor()
        {
          if (block.statements != null)
          {
            for(Statement statement: block.statements)
            {
              refactorStatement(statement);
            }
          }

          addStatement(getBreakLabel(this));
        }
      };
      
      return super.visit(block, scope);
    }
    
    @Override
    public void endVisit(Block block, BlockScope scope) 
    {
      current = current.parent;
    }
    
    @Override
    public boolean visit(
      final LabeledStatement labeledStatement, 
      final BlockScope scope) 
    {
      current = new Scope(current, labeledStatement)
      {
        @Override
        public void refactor()
        {
          refactorStatement(labeledStatement.statement);
        }
      };

      return super.visit(labeledStatement, scope);
    }
    
    @Override
    public void endVisit(LabeledStatement labeledStatement, BlockScope scope)
    {
      current = current.parent;
    }
    
    @Override
    public boolean visit(
      final ForStatement forStatement, 
      final BlockScope scope)
    {
      current = new Scope(current, forStatement)
      {
        @Override
        public void refactor()
        {
          if (forStatement.initializations != null) 
          {
            for(Statement statement: forStatement.initializations)
            {
              refactorStatement(statement);
            }
          }
          
          Label label = label();
          Label breakLabel = getBreakLabel(this);

          addStatement(label);

          if ((forStatement.condition != null) &&
            !(forStatement.condition instanceof TrueLiteral))
          {
            addStatement(
              ifThen(
                new UnaryExpression(forStatement.condition, OperatorIds.NOT),
                block(
                  assign(ref(stateIdName), literal(breakLabel)),
                  doContinue())));                  
          }
          
          refactorStatement(forStatement.action);
          
          addStatement(getIterationLabel(this));
          
          if (forStatement.increments != null)
          {
            for(Statement statement: forStatement.increments)
            {
              refactorStatement(statement);
            }
          }
          
          addStatement(assign(ref(stateIdName), literal(label)));
          addStatement(doContinue());
          addStatement(breakLabel);
        }
      };
      
      return super.visit(forStatement, scope);
    }

    @Override
    public void endVisit(ForStatement forStatement, BlockScope scope) 
    {
      current = current.parent;
    }
    
    @Override
    public boolean visit(
      final ForeachStatement forStatement, 
      final BlockScope scope) 
    {
      current = new Scope(current, forStatement)
      {
        @Override
        public void refactor()
        {
          char[] iteratorVarName = 
            name(
              statePrefix + 
              "$iterator$" + 
              new String(forStatement.elementVariable.name));
          
          FieldDeclaration iterator = 
            fieldDef(iteratorVarName, type(namesJavaUtilIterator));
          
          iterator.annotations = new Annotation[] { suppressWarnings() };
            
          stateVariables.add(iterator);
          
          addStatement( 
            assign(
              ref(iteratorVarName), 
              invoke(null, iteratorName, forStatement.collection)));
          
          addStatement(getIterationLabel(this));
          
          addStatement(
            ifThen(
              new UnaryExpression(
                invoke(ref(iteratorVarName), nameHasNext),
                OperatorIds.NOT), 
              block(
                assign(ref(stateIdName), literal(getBreakLabel(this))),
                doContinue())));
          
          addStatement(
            assign(
              ref(forStatement.elementVariable.name),
              new CastExpression(
                invoke(ref(iteratorVarName), nameNext),
                forStatement.elementVariable.type)));
          
          refactorStatement(forStatement.action);
          
          addStatement(
            assign(ref(stateIdName), literal(getIterationLabel(this))));
          addStatement(doContinue());
          addStatement(getBreakLabel(this));
        }
      };
      
      return super.visit(forStatement, scope);
    }
    
    @Override
    public void endVisit(ForeachStatement forStatement, BlockScope scope) 
    {
      current = current.parent;
    }
    
    @Override
    public boolean visit(
      final DoStatement doStatement, 
      final BlockScope scope) 
    {
      current = new Scope(current, doStatement)
      {
        @Override
        public void refactor()
        {
          addStatement(getIterationLabel(this));
          
          refactorStatement(doStatement.action);

          addStatement(
            ifThen(
              doStatement.condition, 
              block(
                assign(
                  ref(stateIdName), 
                  literal(getIterationLabel(this))),
                doContinue())));
          
          addStatement(getBreakLabel(this));
        }
      };
      
      return super.visit(doStatement, scope);
    }
    
    @Override
    public void endVisit(DoStatement doStatement, BlockScope scope) 
    {
      current = current.parent;
    }
    
    @Override
    public boolean visit(
      final WhileStatement whileStatement, 
      final BlockScope scope) 
    {
      current = new Scope(current, whileStatement)
      {
        @Override
        public void refactor()
        {
          addStatement(getIterationLabel(this));

          if (!(whileStatement.condition instanceof TrueLiteral))
          {
            addStatement(
              ifThen(
                new UnaryExpression(
                  whileStatement.condition, 
                  OperatorIds.NOT),
                block(
                  assign(
                    ref(stateIdName), 
                    literal(getBreakLabel(this))),
                  doContinue())));
          }
          
          refactorStatement(whileStatement.action);
          
          addStatement(
            assign(ref(stateIdName), literal(getIterationLabel(this))));
          addStatement(doContinue());
          addStatement(getBreakLabel(this));
        }
      };
      
      return super.visit(whileStatement, scope);
    }
    
    @Override
    public void endVisit(WhileStatement whileStatement, BlockScope scope) 
    {
      current = current.parent;
    }

    @Override
    public boolean visit(
      final IfStatement ifStatement, 
      final BlockScope scope) 
    {
      current = new Scope(current, ifStatement)
      {
        @Override
        public void refactor()
        {
          Label label = 
            ifStatement.elseStatement == null ? getBreakLabel(this) : label();
          
          if (!(ifStatement.condition instanceof TrueLiteral))
          {
            addStatement(
              ifThen(
                new UnaryExpression(ifStatement.condition, OperatorIds.NOT),
                block(
                  assign(ref(stateIdName), literal(label)), 
                  doContinue())));
          }

          if (ifStatement.elseStatement != null)
          {
            refactorStatement(ifStatement.elseStatement);
            addStatement(
              assign(ref(stateIdName), literal(getBreakLabel(this))));
            addStatement(doContinue());
            addStatement(label);
          }
          
          refactorStatement(ifStatement.thenStatement);
          addStatement(getBreakLabel(this));
        }
      };
      
      return super.visit(ifStatement, scope);
    }
    
    @Override
    public void endVisit(IfStatement ifStatement, BlockScope scope)
    {
      current = current.parent;
    }

    @Override
    public boolean visit(
      final SwitchStatement switchStatement, 
      final BlockScope scope) 
    {
      current = new Scope(current, switchStatement)
      {
        @Override
        public void refactor()
        {
          Label breakLabel = getBreakLabel(this);
          SwitchStatement newSwitchStatement = new SwitchStatement();
          
          newSwitchStatement.expression = switchStatement.expression;
          
          addStatement(newSwitchStatement);
          
          if (switchStatement.statements != null)
          {
            boolean hasDefault = false;
            ArrayList<Statement> cases = new ArrayList<Statement>();

            for(Statement statement: switchStatement.statements)
            {
              if (statement instanceof CaseStatement)
              {
                CaseStatement caseStatement = (CaseStatement)statement;
                
                if (caseStatement.constantExpression == null)
                {
                  hasDefault = true;
                }
                
                Label label = label();
                
                cases.add(caseStatement);
                cases.add(assign(ref(stateIdName), literal(label)));
                cases.add(doContinue());

                addStatement(label);
              }
              else
              {
                refactorStatement(statement);
              }
            }
            
            if (!hasDefault)
            {
              cases.add(new CaseStatement(null, 0, 0));
              cases.add(assign(ref(stateIdName), literal(breakLabel)));
              cases.add(doContinue());
            }

            newSwitchStatement.statements = 
              cases.toArray(new Statement[cases.size()]);
          }
          
          addStatement(breakLabel);
        }
      };
      
      return super.visit(switchStatement, scope);
    }
    
    @Override
    public void endVisit(SwitchStatement switchStatement, BlockScope scope) 
    {
      current = current.parent;
    }
    
    @Override
    public boolean visit(
      final TryStatement tryStatement, 
      final BlockScope scope)
    {
      current = new Scope(current, tryStatement)
      {
        @Override
        public void refactor()
        {
          boolean hasFinally = tryStatement.finallyBlock != null;
          boolean hasCatch = tryStatement.catchBlocks != null;
          ErrorHandler catchHandler = null;
          ErrorHandler finallyHandler = null;
          Label tryLabel = label();
          Label finallyLabel;
          Label breakLabel = getBreakLabel(this);
          char[] finallyErrorName = null;
          
          if (hasFinally)
          {
            finallyHandler = new ErrorHandler();
            finallyLabel = getFinallyLabel(this);
            
            ++finallyBlocks;
            finallyErrorName = name(statePrefix + "$exception" + finallyBlocks);
            labelName = name(statePrefix + "$id" + finallyBlocks);
            
            stateVariables.add(
              fieldDef(finallyErrorName, type(namesJavaLangThrowable)));
            
            stateVariables.add(fieldDef(labelName, type(TypeIds.T_int)));

            addStatement(assign(ref(finallyErrorName), literal()));
            addStatement(assign(ref(labelName), literal(breakLabel)));
          }
          else
          {
            finallyLabel = breakLabel;
          }
          
          addStatement(assign(ref(stateIdName), literal(tryLabel)));
          
          if (hasCatch)
          {
            catchHandler = new ErrorHandler();
            catchHandler.begin = statements.size();
          }
          else 
          {
            if (hasFinally)
            {
              finallyHandler.begin = statements.size();
            }
          }
          
          addStatement(tryLabel);
          refactorStatement(tryStatement.tryBlock);
          addStatement(assign(ref(stateIdName), literal(finallyLabel)));
          
          if (hasCatch)
          {
            addStatement(doContinue());
            catchHandler.end = statements.size();

            for(int i = 0; i < tryStatement.catchBlocks.length; ++i)
            {
              Argument argument = tryStatement.catchArguments[i];
              Block catcher = tryStatement.catchBlocks[i];
              Label label = label();
              
              usedLabels.add(label);
              addStatement(label);
              refactorStatement(catcher);
              addStatement(assign(ref(stateIdName), literal(finallyLabel)));
              addStatement(doContinue());
              
              catchHandler.statements.add(
                ifThen(
                  new InstanceOfExpression(ref(errorName), argument.type),
                  block(
                    assign(
                      ref(argument.name), 
                      new CastExpression(ref(errorName), argument.type)),
                    assign(ref(stateIdName), literal(label)),
                    doContinue())));
            }

            errorHandlers.add(catchHandler);
            
            if (hasFinally)
            {
              finallyHandler.begin = catchHandler.end;
            }
          }
          
          if (hasFinally)
          {
            finallyHandler.end = statements.size();

            addStatement(finallyLabel);
            refactorStatement(tryStatement.finallyBlock);
            
            addStatement(
              ifThen(
                new EqualExpression(
                  ref(finallyErrorName), 
                  literal(), 
                  OperatorIds.NOT_EQUAL),
                block(
                  assign(ref(errorName), ref(finallyErrorName)),
                  new BreakStatement(null, 0, 0))));

            Scope next = getFinallyScope(parent, null);
            
            if (next != null)
            {
              Label label = getFinallyLabel(next);
              
              addStatement(
                new IfStatement(
                  source(
                    new BinaryExpression(
                      ref(labelName), 
                      literal(label),
                      OperatorIds.GREATER)), 
                  block(
                    assign(ref(next.labelName), ref(labelName)),
                    assign(ref(stateIdName), literal(label))), 
                  assign(ref(stateIdName), ref(labelName)),
                  0,
                  0));
            }
            else
            {
              addStatement(assign(ref(stateIdName), ref(labelName)));
            }
            
            addStatement(doContinue());

            finallyHandler.statements.add(
              assign(ref(finallyErrorName), ref(errorName)));
            finallyHandler.statements.add(
              assign(ref(stateIdName), literal(finallyLabel)));
            finallyHandler.statements.add(doContinue());
            
            usedLabels.add(finallyLabel);
            errorHandlers.add(finallyHandler);
          }
          
          addStatement(breakLabel);
        }
      };
      
      return super.visit(tryStatement, scope);
    }
    
    @Override
    public void endVisit(TryStatement tryStatement, BlockScope scope) 
    {
      current = current.parent;
    }
    
    @Override
    public boolean visit(
      final LocalDeclaration localDeclaration, 
      final BlockScope scope)
    {
      if (localDeclaration == iteratorDecl)
      {
        return false;
      }

      if (equal(localDeclaration.name, iteratorDecl.name))
      {
        invalid("No variable with the same name as iterable is permitted.");
      }

      variableDecls.add(
        new Scope(current, localDeclaration)
        {
          @Override
          public void refactor()
          {
            if (localDeclaration.initialization != null)
            {
              if (localDeclaration.initialization instanceof ArrayInitializer)
              {
                ArrayInitializer initializer = 
                  (ArrayInitializer)localDeclaration.initialization;
                ArrayAllocationExpression allocation = 
                  new ArrayAllocationExpression();
                
                TypeReference type = localDeclaration.type; 
                
                if (type instanceof SingleTypeReference)
                {
                  SingleTypeReference reference = (SingleTypeReference)type;

                  allocation.type = type(reference.token);
                }
                else if (type instanceof QualifiedTypeReference)
                {
                  QualifiedTypeReference reference = 
                    (QualifiedTypeReference)type;

                  allocation.type = type(reference.tokens);
                }
                
                allocation.initializer = initializer;
                allocation.dimensions = 
                  new Expression[localDeclaration.type.dimensions()];
                
                addStatement(assign(ref(localDeclaration.name), allocation));
              }
              else
              {
                addStatement(
                  assign(
                    ref(localDeclaration.name), 
                    localDeclaration.initialization));
              }
            }
          }
        });
      
      return super.visit(localDeclaration, scope);
    }
    
    @Override
    public boolean visit(
      final Argument argument, 
      final BlockScope scope) 
    {
      current = new Scope(current, argument)
      {
        @Override
        public void refactor() { }
      };

      if (!(current.parent.node instanceof MethodDeclaration))
      {
        variableDecls.add(current);
      }
      
      return super.visit(argument, scope);
    }
    
    @Override
    public void endVisit(Argument argument, BlockScope scope) 
    {
      current = current.parent;
    }

    @Override
    public boolean visit(
      final ReturnStatement returnStatement, 
      final BlockScope scope)
    {
      if (!validateReturn(returnStatement))
      {
        invalid(
          iterable ?
            "Return, if any, should be of the form: return items;" :
            "Return, if any, should be of the form: return items.iterate();");
      }
      
      current = new Scope(current, returnStatement)
      {
        @Override
        public void refactor()
        {
          Label label = getBreakLabel(root);
          Scope next = getFinallyScope(parent, null);
          
          if (next == null)
          {
            addStatement(assign(ref(stateIdName), literal(label)));
            addStatement(doContinue());
          }
          else
          {
            addStatement(assign(ref(next.labelName), literal(label)));
            addStatement(
              assign(ref(stateIdName), literal(getFinallyLabel(next))));
            addStatement(doContinue());
          }
        }
      };
      
      breaks.add(current);
      
      return false;
    }
    
    @Override
    public void endVisit(ReturnStatement returnStatement, BlockScope scope) 
    {
      current = current.parent;
    }
    
    @Override
    public boolean visit(
      final BreakStatement breakStatement, 
      final BlockScope scope)
    {
      char[] label = breakStatement.label;
      Scope target = null;
      
      if (label != null)
      {
        Scope labelScope = current;
        
        while(labelScope != null)
        {
          if (labelScope.node instanceof LabeledStatement)
          {
            LabeledStatement labeledStatement = 
              (LabeledStatement)labelScope.node;
            
            if (equal(label, labeledStatement.label))
            {
              if (target != null)
              {
                invalid("Invalid label.");  
              }
              
              target = labelScope;
            }
          }

          labelScope = labelScope.parent;
        }
      }
      else
      {
        Scope labelScope = current;
        
        while(labelScope != null)
        {
          if ((labelScope.node instanceof ForStatement) ||
            (labelScope.node instanceof ForeachStatement) ||
            (labelScope.node instanceof WhileStatement) ||
            (labelScope.node instanceof DoStatement) ||
            (labelScope.node instanceof SwitchStatement))
          {
            target = labelScope;
            
            break;
          }

          labelScope = labelScope.parent;
        }
      }
      
      if (target == null)
      {
        invalid("Invalid break.");
      }
      
      current = new Scope(current, breakStatement)
      {
        @Override
        public void refactor()
        {
          Scope next = getFinallyScope(parent, target);
          Label label = getBreakLabel(target);
          
          if (next == null)
          {
            addStatement(assign(ref(stateIdName), literal(label)));
            addStatement(doContinue());
          }
          else
          {
            addStatement(assign(ref(next.labelName), literal(label)));
            addStatement(
                assign(ref(stateIdName), literal(getFinallyLabel(next))));
            addStatement(doContinue());
          }
        }
      };
      
      current.target = target;
      breaks.add(current);
      
      return false;
    }
    
    @Override
    public void endVisit(BreakStatement breakStatement, BlockScope scope)
    {
      current = current.parent;
    }
    
    @Override
    public boolean visit(
      final ContinueStatement continueStatement, 
      final BlockScope scope)
    {
      char[] label = continueStatement.label;
      Scope target = null;
      
      if (label != null)
      {
        Scope labelScope = current;
        
        while(labelScope != null)
        {
          if (labelScope.node instanceof LabeledStatement)
          {
            LabeledStatement labeledStatement =
              (LabeledStatement)labelScope.node;
            
            if (label == labeledStatement.label)
            {
              if (target != null)
              {
                invalid("Invalid label.");  
              }
              
              if ((labelScope.node instanceof ForStatement) ||
                (labelScope.node instanceof ForeachStatement) ||
                (labelScope.node instanceof WhileStatement) ||
                (labelScope.node instanceof DoStatement))
              {
                target = labelScope;
              }
              else
              {
                invalid("Invalid continue.");
              }
            }
          }
          
          labelScope = labelScope.parent;
        }
      }
      else
      {
        Scope labelScope = current;
        
        while(labelScope != null)
        {
          if ((labelScope.node instanceof ForStatement) ||
            (labelScope.node instanceof ForeachStatement) ||
            (labelScope.node instanceof WhileStatement) ||
            (labelScope.node instanceof DoStatement))
          {
            target = labelScope;
            
            break;
          }

          labelScope = labelScope.parent;
        }
      }
      
      if (target == null)
      {
        invalid("Invalid continue.");
      }
      
      current = new Scope(current, continueStatement)
      {
        @Override
        public void refactor()
        {
          Scope next = getFinallyScope(parent, target);
          Label label = getIterationLabel(target);

          if (next == null)
          {
            addStatement(assign(ref(stateIdName), literal(label)));
            addStatement(doContinue());
          }
          else
          {
            addStatement(assign(ref(next.labelName), literal(label)));
            addStatement(
                assign(ref(stateIdName), literal(getFinallyLabel(next))));
            addStatement(doContinue());
          }
        }
      };
        
      current.target = target;
      breaks.add(current);
      
      return false;
    }
    
    @Override
    public void endVisit(ContinueStatement continueStatement, BlockScope scope)
    {
      current = current.parent;
    }
    
    @Override
    public boolean visit(ThisReference thisReference, BlockScope scope) 
    {
      if (!thisReference.isImplicitThis())
      {
        invalid("No unqualified \"this\" expression is permitted.");
      }
      
      return false;
    }
    
    @Override
    public boolean visit(SuperReference thisReference, BlockScope scope) 
    {
      invalid("No unqualified \"super\" expression is permitted.");
      
      return false;
    }

    @Override
    public boolean visit(
      SingleNameReference singleNameReference, 
      BlockScope scope)
    {
      if (equal(singleNameReference.token, iteratorDecl.name))
      {
        invalid("No use of iterable is permitted in contexts other " +
          "than defined by the yield pattern.");
      }
      
      names.add(new String(singleNameReference.token));
      
      return super.visit(singleNameReference, scope);
    }

    @Override
    public boolean visit(
      TypeDeclaration localTypeDeclaration, 
      BlockScope scope) 
    {
      if (localTypeDeclaration.allocation == null)
      {
        classes.add(localTypeDeclaration);
      }
      
      return false;
    }
    
    @Override
    public boolean visit(
      final MessageSend messageSend, 
      final BlockScope scope)
    {
      if (messageSend.receiver.isImplicitThis())
      {
        char[] name = messageSend.selector;
        
        if ((iterable && equal(name, nameIterator)) ||
          equal(name, nameHasNext) ||
          equal(name, nameNext) ||
          equal(name, nameRemove) ||
          equal(name, nameClose))
        {
          invalid(
            "Cannot call method " + new String(name) + "(), as it is hidden.");
        }
        
        return super.visit(messageSend, scope);
      }
      
      final Expression expression = getYieldExpression(messageSend);
      
      if (expression != null)
      {
        yields.add(
          new Scope(current, messageSend)
          {
            @Override
            public void refactor()
            {
              Label label = getBreakLabel(this);
              
              addStatement(assign(ref(nextName), expression));
              addStatement(assign(ref(stateIdName), literal(label)));
              addStatement(ret(literal(true)));
              addStatement(label);
              
              Scope next = getFinallyScope(parent, null);
              
              if (next != null)
              {
                breakCases.add(source(new CaseStatement(literal(label), 0, 0)));
                breakCases.add(
                  assign(ref(next.labelName), literal(getBreakLabel(root))));
                breakCases.add(
                  assign(ref(stateIdName), literal(getFinallyLabel(next))));
                breakCases.add(doContinue());
              }
            }
          });
        
        expression.traverse(this, scope);
        
        return false;
      }
      else
      {
        return super.visit(messageSend, scope);
      }
    }
  }
}