package com.nesterovskyBros.annotation.processor.javac;

import java.util.Set;

import javax.annotation.processing.ProcessingEnvironment;
import javax.annotation.processing.RoundEnvironment;

import javax.lang.model.element.ExecutableElement;
import javax.tools.Diagnostic.Kind;

import com.sun.tools.javac.processing.JavacProcessingEnvironment;

/**
 * A transformer factory for the yield processor.
 */
public class YieldService 
  implements com.nesterovskyBros.annotation.processor.YieldService 
{
  /**
   * {@inheritDoc}
   */
  public boolean supports(
    ProcessingEnvironment processingEnv,
    RoundEnvironment roundEnv)
  {
    try
    {
      Class<?> type = Class.forName(
        "com.sun.tools.javac.processing.JavacProcessingEnvironment");
      
      return type.cast(processingEnv) != null;
    }
    catch(Throwable e)
    {
      return false;
    }
  }
  
  /**
   * Transforms methods.
   * @param methods methods to transform.
   * @param processingEnv a {@link ProcessingEnvironment} instance.
   * @param roundEnv a {@link RoundEnvironment} instance.
   * @return true if operation succeeds, and false otherwise.
   */
  public boolean transform(
    ProcessingEnvironment processingEnv,
    RoundEnvironment roundEnv,
    Set<ExecutableElement> methods)
  {
    return Runner.transform(processingEnv, roundEnv, methods);
  }
  
  private static class Runner
  {
    /**
     * Transforms methods.
     * @param methods methods to transform.
     * @param processingEnv a {@link ProcessingEnvironment} instance.
     * @param roundEnv a {@link RoundEnvironment} instance.
     * @return true if operation succeeds, and false otherwise.
     */
    public static boolean transform(
      ProcessingEnvironment processingEnv,
      RoundEnvironment roundEnv,
      Set<ExecutableElement> methods)
    {
      JavacProcessingEnvironment environment = 
        (JavacProcessingEnvironment)processingEnv;

      boolean debug = environment.getOptions().containsKey("debug");
      boolean succeed = false;
      
      for(ExecutableElement method: methods)
      {
        YieldTransformer transformer = 
          new YieldTransformer(environment, method);
        
        succeed |= transformer.transform();
        
        if (debug)
        {
          environment.getMessager().printMessage(
            Kind.NOTE, 
            transformer.methodTree.toString());
        }
      }
      
      return succeed;
    }
  }
}
