package com.nesterovskyBros.annotation.processor.javac;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import javax.lang.model.element.Element;
import javax.lang.model.element.ExecutableElement;

import javax.lang.model.util.Elements;
import javax.lang.model.util.Types;

import javax.tools.Diagnostic.Kind;

import com.sun.source.util.Trees;

import com.sun.tools.javac.code.Flags;
import com.sun.tools.javac.code.TypeTags;

import com.sun.tools.javac.processing.JavacProcessingEnvironment;

import com.sun.tools.javac.tree.JCTree;
import com.sun.tools.javac.tree.TreeCopier;
import com.sun.tools.javac.tree.TreeInfo;
import com.sun.tools.javac.tree.TreeMaker;
import com.sun.tools.javac.tree.TreeScanner;

import com.sun.tools.javac.tree.JCTree.JCAnnotation;
import com.sun.tools.javac.tree.JCTree.JCAssign;
import com.sun.tools.javac.tree.JCTree.JCBlock;
import com.sun.tools.javac.tree.JCTree.JCBreak;
import com.sun.tools.javac.tree.JCTree.JCCase;
import com.sun.tools.javac.tree.JCTree.JCCatch;
import com.sun.tools.javac.tree.JCTree.JCClassDecl;
import com.sun.tools.javac.tree.JCTree.JCContinue;
import com.sun.tools.javac.tree.JCTree.JCDoWhileLoop;
import com.sun.tools.javac.tree.JCTree.JCEnhancedForLoop;
import com.sun.tools.javac.tree.JCTree.JCExpression;
import com.sun.tools.javac.tree.JCTree.JCExpressionStatement;
import com.sun.tools.javac.tree.JCTree.JCForLoop;
import com.sun.tools.javac.tree.JCTree.JCIdent;
import com.sun.tools.javac.tree.JCTree.JCIf;
import com.sun.tools.javac.tree.JCTree.JCLabeledStatement;
import com.sun.tools.javac.tree.JCTree.JCLiteral;
import com.sun.tools.javac.tree.JCTree.JCMethodDecl;
import com.sun.tools.javac.tree.JCTree.JCMethodInvocation;
import com.sun.tools.javac.tree.JCTree.JCNewClass;
import com.sun.tools.javac.tree.JCTree.JCReturn;
import com.sun.tools.javac.tree.JCTree.JCSkip;
import com.sun.tools.javac.tree.JCTree.JCStatement;
import com.sun.tools.javac.tree.JCTree.JCSwitch;
import com.sun.tools.javac.tree.JCTree.JCSynchronized;
import com.sun.tools.javac.tree.JCTree.JCTry;
import com.sun.tools.javac.tree.JCTree.JCTypeApply;
import com.sun.tools.javac.tree.JCTree.JCTypeParameter;
import com.sun.tools.javac.tree.JCTree.JCVariableDecl;
import com.sun.tools.javac.tree.JCTree.JCWhileLoop;

import com.sun.tools.javac.util.Context;
import com.sun.tools.javac.util.List;
import com.sun.tools.javac.util.ListBuffer;
import com.sun.tools.javac.util.Name;

/**
 * A yield method transformer.
 */
public class YieldTransformer 
{
  /**
   * Creates a validator instance.
   * @param environment a processing environment.
   * @param element a method element.
   */
  public YieldTransformer(
    JavacProcessingEnvironment environment,
    ExecutableElement element)
  {
    this.environment = environment;
    this.method = element;

    Context context = environment.getContext();

    types = environment.getTypeUtils();
    elements = environment.getElementUtils();
    trees = Trees.instance(environment);
    maker = TreeMaker.instance(context);
    copier = new TreeCopier<JCTree>(maker);
    methodTree = (JCMethodDecl)trees.getTree(method);
    
    _this = name("this");
    _super = name("super");
    empty = name("");
  }
  
  /**
   * Transforms the method body.
   * @return true if operation succeeds, and false otherwise.
   */
  public boolean transform()
  {
    try
    {
      if (!validate())
      {
        return false;
      }
    }
    catch(Exception e) 
    {
      return false;
    }
    
    return refactor();
  }
  
  /**
   * Issue an error and throws exception. 
   * @param message an error message.
   */
  private void invalid(String message)
  {
    environment.getMessager().printMessage(Kind.ERROR, message, method);
    
    throw new IllegalStateException(message);
  }
  
  /**
   * Validates assumptions.
   * @return true if refactoring is possible, and false otherwise.
   */
  private boolean validate()
  {
    if ((methodTree.mods != null) && 
      ((methodTree.mods.flags & Flags.SYNCHRONIZED) != 0))
    {
      invalid("Method should be not synchronized.");
    }

    if (!isEmpty(methodTree.params))
    {
      for(JCVariableDecl param: methodTree.params)
      {
        if ((param.mods == null) || (param.mods.flags & Flags.FINAL) == 0)
        {
          invalid("Parameters should be final.");
        }
      }
    }
    
    Element type = types.asElement(method.getReturnType());
    
    if (type.equals(elements.getTypeElement("java.lang.Iterable")))
    {
      iterable = true;
    }
    else if (type.equals(elements.getTypeElement("java.util.Iterator")))
    {
      iterable = false;
    }
    else
    {
      invalid("Return type should be either Iterable or Iterator.");
    }
    
    if (methodTree.body == null)
    {
      invalid("Method body is required.");
    }
    
    if (!isEmpty(methodTree.body.stats))
    {
      for(JCStatement statement: methodTree.body.stats)
      {
        if ((statement instanceof JCClassDecl) ||
          (statement instanceof JCSkip))
        {
          continue;
        }

        if (statement instanceof JCVariableDecl)
        {
          iteratorDecl = (JCVariableDecl)statement;
          
          if (!isArrayList(iteratorDecl.vartype) || 
            !isNewArrayList(iteratorDecl.init))
          {
            iteratorDecl = null;
          }
        }

        break;
      }
    }
    
    if (iteratorDecl == null)
    {
      invalid("The first statement should be of the form " +
        "ArrayList<T> items = new ArrayList<T>();");
    }
    
    ValidationScanner scanner = new ValidationScanner();

    scanner.scan(methodTree.body);
    
    if (yields.isEmpty())
    {
      invalid("No yield statements are found in the form: items.add(...);");
    }
    
    for(Scope scope: yields)
    {
      Scope yieldScope = scope;
      
      do
      {
        allScopes.put(yieldScope.node, yieldScope);
        yieldScope = yieldScope.parent;
      }
      while(yieldScope != null);
    }

    // Collect scope from breaks.
    boolean collected = !breaks.isEmpty();

    while(collected)
    {
      collected = false;
      
      for(Scope scope: breaks)
      {
        Scope target = scope.target;
        
        if (((target == null) || allScopes.containsKey(target.node)) && 
          !allScopes.containsKey(scope.node))
        {
          collected = true;
          
          Scope breakScope = scope;
          
          do
          {
            allScopes.put(breakScope.node, breakScope);
            breakScope = breakScope.parent;
          }
          while(breakScope != null);
        }
      }
    }
    
    for(Scope scope: variableDecls)
    {
      boolean stateVariable = false;
      
      if (allScopes.containsKey(scope.parent.node))
      {
        stateVariable = true;
      }
      else 
      {
        if ((scope.parent.node instanceof JCCatch) &&
          allScopes.containsKey(scope.parent.parent.node))
        {
          stateVariable = true;
        }
      }
      
      if (stateVariable)
      {
        JCVariableDecl variable = (JCVariableDecl)scope.node;
        
        allScopes.put(scope.node, scope);
        
        stateVariables.append(
          varDef(Flags.PRIVATE, variable.name, copy(variable.vartype)));
      }
    }
    
    root = allScopes.get(methodTree.body);
    
    for(JCTree tree: allScopes.keySet())
    {
      if (tree instanceof JCSynchronized)
      {
        invalid("No synchronized statement is allowed in this context.");
      }
      
      if (tree instanceof JCEnhancedForLoop)
      {
        createIterator = true;
      }
    }

    if (!stateVariables.isEmpty())
    {
      HashSet<Name> names = new HashSet<Name>();
      
      for(JCVariableDecl variable: stateVariables)
      {
        if (!names.add(variable.name))
        {
          invalid("No variables with duplicate names " +
            "should be used in the state machine.");
        }
      }
    }
    
    JCTypeApply returnMethodType = (JCTypeApply)methodTree.restype;
    
    itemType = returnMethodType.arguments.head;
    
    statePrefix = "$state";

Prefix:
    for(int i = 1;; ++i)
    {
      for(String name: names)
      {
        if (name.startsWith(statePrefix))
        {
          statePrefix = "$state" + i;
          
          continue Prefix;
        }
      }
  
      break;
    }
  
    stateIdName = name(statePrefix + "$id");
    hasNextName = name(statePrefix + "$hasNext");
    nextDefinedName = name(statePrefix + "$nextDefined");
    nextName = name(statePrefix + "$next");
    errorName = name(statePrefix + "$exception");
    moveNextName = name(statePrefix + "$next");
    
    if (createIterator)
    {
      iteratorName = name(statePrefix + "$iterator");
    }

    stateVariables.append(
      varDef(Flags.PRIVATE, stateIdName, type(TypeTags.INT)));
    stateVariables.append(
      varDef(Flags.PRIVATE, hasNextName, type(TypeTags.BOOLEAN)));
    stateVariables.append(
      varDef(Flags.PRIVATE, nextDefinedName, type(TypeTags.BOOLEAN)));
    stateVariables.append(varDef(Flags.PRIVATE, nextName, copy(itemType)));

    return true;
  }

  /**
   * Transforms method body.
   */
  private boolean refactor()
  {
    current = root;
    
    JCCase begin = label();
    JCCase iteratorLabel = getIterationLabel(root);
    
    usedLabels.add(begin);
    usedLabels.add(iteratorLabel);
    usedLabels.add(getBreakLabel(root));

    handleRewrites();
    addStatement(begin);
    addStatement(assign(stateIdName, literal(iteratorLabel)));
    addStatement(iteratorLabel);
    root.refactor();
    endCase();
    
    optimizeBreaks();
    modifyMethod();
    
    return true;
  }
  
  /**
   * Copies a tree.
   * @param <T> a tree type.
   * @param tree a tree to copy.
   * @return a copy of the tree.
   */
  private <T extends JCTree> T copy(T tree) 
  {
    return copier.copy(tree);
  }

  /**
   * Gets a tree maker instance initialized to the current position.
   */
  private TreeMaker maker()
  {
    return maker.at(current != null ? current.node.pos : 0);
  }

  /**
   * Returns true if list is null or empty.
   * @param <T> an element type.
   * @param list a list to check.
   * @return true if list is null or empty, and false otherwise.
   */
  private static <T> boolean isEmpty(List<T> list)
  {
    return (list == null) || list.isEmpty();
  }
  
  /**
   * Tests whether expression represents ArrayList type.
   * @param expression an expression to test.
   * @return true if expression represents ArrayList type, and 
   *   false otherwise.
   */
  private boolean isArrayList(JCExpression expression)
  {
    if (expression instanceof JCTypeApply)
    {
      JCTypeApply type = (JCTypeApply)expression;
      Name name = TreeInfo.fullName(type.clazz);
      
      return !isEmpty(type.arguments) &&
        (type.arguments.size() == 1) &&
        (type.clazz != null) &&
        ((name("java.util.ArrayList") == name) || (name("ArrayList") == name));
    }
    
    return false;
  }

  /**
   * Tests whether expression is of the form: ArrayList<T>().
   * @param expression an expression to test.
   * @return true if exression of the form ArrayList<T>(), and 
   *   false otherwise.
   */
  private boolean isNewArrayList(JCExpression expression)
  {
    expression = TreeInfo.skipParens(expression);
    
    if (expression instanceof JCNewClass)
    {
      JCNewClass newClass = (JCNewClass)expression;
      
      return (newClass.encl == null) &&
        isEmpty(newClass.typeargs) &&
        isEmpty(newClass.args) &&
        (newClass.def == null) &&
        isArrayList(newClass.clazz);
    }
    
    return false;
  }
  
  /**
   * Validates return statement.
   * @param expression an expression to test.
   * @return true if it's a valid return statement, and false otherwise.
   */
  private boolean validateReturn(JCReturn statement)
  {
    JCExpression expression = TreeInfo.skipParens(statement.expr);
    
    if (iterable)
    {
      return (expression instanceof JCIdent) &&
        (iteratorDecl.name == TreeInfo.name(expression));
    }
    else
    {
      if (expression instanceof JCMethodInvocation)
      {
        JCMethodInvocation invoke = (JCMethodInvocation)expression;

        return isEmpty(invoke.typeargs) &&
          isEmpty(invoke.args) &&
          (name(iteratorDecl.name + ".iterator") == 
            TreeInfo.fullName(invoke.meth));            
      }
    }
    
    return false;
  }
  
  /**
   * Gets yield expression from yield statement: iterator.add(...);
   * @param statement a statement.
   * @return a yield expression, or null if it's not a yield statement.
   */
  private JCExpression getYieldExpression(JCExpression expression)
  {
    if (expression instanceof JCMethodInvocation)
    {
      JCMethodInvocation invoke = (JCMethodInvocation)expression;

      if (isEmpty(invoke.typeargs) &&
        !isEmpty(invoke.args) && 
        (invoke.args.size() == 1) &&
        (name(iteratorDecl.name + ".add") == TreeInfo.fullName(invoke.meth)))
      {
        return invoke.args.head;
      }
    }
    
    return null;
  }
  
  /**
   * Gets set stated id, if any.
   * @param statement a statement to test.
   * @return label expression, or null if statement 
   *   is not of the specified form.
   */
  private JCExpression getSetStateId(JCStatement statement)
  {
    if (statement instanceof JCExpressionStatement)
    {
      JCExpressionStatement exec = (JCExpressionStatement)statement;
      
      if (exec.expr instanceof JCAssign)
      {
        JCAssign assign = (JCAssign)exec.expr;
        
        if (assign.lhs instanceof JCIdent)
        {
          JCIdent ident = (JCIdent)assign.lhs;
          
          if (ident.name == stateIdName)
          {
            return assign.rhs;
          }
        }
      }
    }
    
    return null;
  }
  
  /**
   * Tests whether the statement is continue;
   * @param statement a statement to test.
   * @return true if statement is continue, and false otherwise.
   */
  private boolean isContinue(JCStatement statement)
  {
    if (statement instanceof JCContinue)
    {
      JCContinue continueStatement = (JCContinue)statement;
      
      if (continueStatement.label == null)
      {
        return true;
      }
    }
    
    return false;
  }
  
  
  /**
   * Create a continue label.
   * @return a continue label.
   */
  private JCContinue doContinue()
  {
    return maker().Continue(null);    
  }

  /**
   * Tests whether the value is a "true" literal.
   * @param expression an expression to test.
   * @return true if the value is a "true" literal, and false otherwise.
   */
  private boolean isTrueLiteral(JCExpression expression)
  {
    if (expression instanceof JCLiteral)
    {
      JCLiteral literal = (JCLiteral)expression;
      
      if ((literal.typetag == TypeTags.BOOLEAN) &&
        Boolean.TRUE.equals(literal.getValue()))
      {
        return true;
      }
    }
    
    return false;
  }
  
  /**
   * Gets a label from label expression.
   * @param expression a label expression.
   * @return a label or null if expression is not literal label.
   */
  private JCCase getLabel(JCExpression expression)
  {
    if (expression instanceof JCLiteral)
    {
      JCLiteral literal = (JCLiteral)expression;
      
      return labelLiterals.get(literal);
    }
    
    return null;
  }
  
  /**
   * Creates a new label.
   * @return a new label.
   */
  private JCCase label()
  {
    return maker().Case(null, List.<JCStatement>nil());
  }
  
  /**
   * Ends current case.
   */
  private void endCase()
  {
    if (!cases.isEmpty())
    {
      JCCase lastCase = cases.get(cases.size() - 1);
      
      if (isEmpty(lastCase.stats) && !statements.isEmpty())
      {
        lastCase.stats = statements.toList();
        statements.clear();
      }
    }
  }

  /**
   * Adds a label statement.
   * @param label a label statement.
   */
  private void addStatement(JCCase label)
  {
    endCase();
    label.pat = literal(cases.size());
    
    if (current != null)
    {
      label.pos = current.node.pos;
    }
    
    cases.add(label);
  }
  
  /**
   * Adds a statement to a current case.
   * @param statement a case statement.
   */
  private void addStatement(JCStatement statement)
  {
    statements.append(statement);
  }
  
  /**
   * Gets a break label.
   * @param scope a scope to get break label for.
   * @return a break label.
   */
  private JCCase getBreakLabel(Scope scope)
  {
    JCCase label = scope.breakLabel;
    
    if (label == null)
    {
      label = label();
      scope.breakLabel = label;
    }
    
    return label;
  }

  /**
   * Gets a break label.
   * @param scope a scope to get finally label for.
   * @return a break label.
   */
  private JCCase getFinallyLabel(Scope scope)
  {
    JCCase label = scope.finallyLabel;
    
    if (label == null)
    {
      label = label();
      scope.finallyLabel = label;
    }
    
    return label;
  }

  /**
   * Gets an iteration label.
   * @param scope a scope to get iteration label for.
   * @return an iteration label.
   */
  private JCCase getIterationLabel(Scope scope)
  {
    JCCase label = scope.iterationLabel;
    
    if (label == null)
    {
      label = label();
      scope.iterationLabel = label;
    }
    
    return label;
  }


  /**
   * Gets {@link Name} instance for a name.
   * @param value a name value.
   * @return a {@link Name} instance.
   */
  private Name name(String value)
  {
    Name name = namesCache.get(value);
    
    if (name == null)
    {
      name = (Name)elements.getName(value);
    }
    
    return name;
  }

  /**
   * Creates {@link JCExpression} for a name.
   * @param names names to convert into expression.
   * @return a {@link JCExpression} instance.
   */
  private JCExpression ident(String name)
  {
    JCExpression expression = null;
    int i = 0;
    
    do
    {
      int p = name.indexOf('.', i);
      Name value = name(p == -1 ? name.substring(i) : name.substring(i, p));
      
      expression = expression == null ? ident(value) : 
        maker().Select(expression, value); 
      i = p + 1;
    }
    while(i > 0);
    
    return expression;
  }
  
  /**
   * Creates {@link JCExpression} for a name.
   * @param names names to convert into expression.
   * @return a {@link JCExpression} instance.
   */
  private JCExpression ident(Name name)
  {
    return maker().Ident(name);
  }
  
  /**
   * Creates a type expression.
   * @param name a type name.
   * @return a type expression.
   */
  private JCExpression type(String name)
  {
    return ident(name);
  }
  
  /**
   * Creates a basic type expression.
   * @param typeId a type id.
   * @return a basic type expression.
   */
  private JCExpression type(int typeId)
  {
    return maker().TypeIdent(typeId); 
  }

  /**
   * Creates assignment statement.
   * @param variable a variable to assign to.
   * @param expression an expression to assign.
   * @return an assignment statement.
   */
  private JCStatement assign(Name variable, JCExpression expression)
  {
    return maker().Exec(maker().Assign(ident(variable), expression));
  }

  /**
   * Gets a list from array.
   * @param <T> a list type.
   * @param items an array of items.
   * @return a list instance.
   */
  @SuppressWarnings("unchecked")
  private <T> List<T> list(T... items)
  {
    return List.from(items);
  }

  /**
   * Creates a block from statement.
   * @param statements list of statement.
   * @return a block instance.
   */
  private JCBlock block(JCStatement... statements)
  {
    return maker().Block(0, list(statements));
  }
  
  /**
   * Creates a return statement.
   * @param expression a return value.
   * @return a return statement.
   */
  private JCStatement ret(JCExpression expression)
  {
    return maker().Return(expression);
  }

  /**
   * Creates if-then statement.
   * @param condition a condition expression.
   * @param statement a then statement.
   * @return {@link JCIf} statement.
   */
  private JCStatement ifThen(JCExpression condition, JCStatement then)
  {
    return maker().If(condition, then, null);
  }


  /**
   * Creates new Class() expression.
   * @param type a class type.
   * @param arguments arguments.
   * @return a new Class() expression.
   */
  private JCNewClass newClass(JCExpression type, JCExpression... arguments)
  {
    return maker().NewClass(
      null, 
      List.<JCExpression>nil(), 
      type, 
      list(arguments), 
      null);
  }

  /**
   * Creates a method definition.
   * @param flags method flags.
   * @param name method name.
   * @param body method body.
   * @param restype result type.
   * @param params method parameters.
   * @return a method definition.
   */
  private JCMethodDecl methodDef(
    long flags,
    Name name,
    JCBlock body,
    JCExpression restype,
    JCVariableDecl... params)
  {
    return maker().MethodDef(
      maker().Modifiers(flags), 
      name, 
      restype, 
      List.<JCTypeParameter>nil(), 
      list(params), 
      List.<JCExpression>nil(), 
      body, 
      null);
  }
  
  /**
   * Creates a variable or field definition.
   * @param flags a variable flags.
   * @param name a variable name.
   * @param type a variable type.
   * @return a variable definition.
   */
  private JCVariableDecl varDef(long flags, Name name, JCExpression type)
  {
    return maker().VarDef(maker().Modifiers(flags), name, type, null);
  }
  
  /**
   * Creates an integer literal.
   * @param value a literal value.
   * @return an integer literal.
   */
  private JCLiteral literal(int value)
  {
    return maker().Literal(value);
  }

  /**
   * Creates a boolean literal.
   * @param value a literal value.
   * @return a boolean literal.
   */
  private JCLiteral literal(boolean value)
  {
    return maker().Literal(TypeTags.BOOLEAN, value ? 1 : 0);
  }

  /**
   * Creates a null literal.
   * @return a null literal.
   */
  private JCLiteral literal()
  {
    return maker().Literal(TypeTags.BOT, null);
  }

  /**
   * Creates a label literal. 
   * @param label a label reference.
   */
  private JCLiteral literal(JCCase label)
  {
    JCLiteral literal = label.pat != null ? 
      copy((JCLiteral)label.pat) : literal(-1);
    
    labelLiterals.put(literal, label);
    
    return literal;
  }
  
  /**
   * Gets closest try scope with finally clause, if any.
   * @param scope a start scope.
   * @param top optional top scope.
   * @return closest try scope with finally clause, if any.
   */
  private static Scope getFinallyScope(Scope scope, Scope top)
  {
    JCTree previous = null;
    
    while(scope != null)
    {
      JCTree tree = scope.node;
      
      if (tree instanceof JCTry)
      {
        JCTry statement = (JCTry)tree;
        
        if ((statement.finalizer != null) && (statement.finalizer != previous))
        {
          return scope;
        }
      }
      
      if (scope == top)
      {
        break;
      }
      
      previous = tree;
      scope = scope.parent;
    }
    
    return null;
  }
  
  /**
   * Invokes a method.
   * @param instance a method instance.
   * @param name a method name.
   * @param arguments method arguments.
   * @return an invoke expression.
   */
  private JCExpression invoke(
    JCExpression instance, 
    Name name, 
    JCExpression... arguments)
  {
    return maker().Apply(
      List.<JCExpression>nil(),
      instance == null ? ident(name) : maker().Select(instance, name), 
      list(arguments));
  }
  
  /**
   * Creates an annotation to suppress all warnings.
   * @return An annotation to suppress all warnings.
   */
  private JCAnnotation suppressWarnings()
  {
    JCExpression all = maker().Literal("all");
    
    return maker().Annotation(type("java.lang.SuppressWarnings"), list(all)); 
  }

  /**
   * Refactors a statement.
   * @param statement a statement to refactor.
   */
  private void refactorStatement(JCStatement statement)
  {
    if (statement == null)
    {
      return;
    }

    Scope scope = allScopes.get(statement);
    
    if (scope != null)
    {
      Scope previous = current;
      
      current = scope;
      scope.refactor();
      current = previous;
    }
    else
    {
      if (statement != iteratorDecl)
      {
        addStatement(copy(statement));        
      }
    }
  }
  
  /**
   * Rewrites code.
   */
  private void handleRewrites()
  {
    for(Scope scope: rewrites)
    {
      scope.refactor();
    }
  }
  
  /**
   * Optimizes break statements.
   */
  private void optimizeBreaks()
  {
    int count = cases.size();
    
    for(Map.Entry<JCLiteral, JCCase> entry: labelLiterals.entrySet())
    {
      JCCase label = entry.getValue();
      
      while(label.pat != null)
      {
        if (isEmpty(label.stats))
        {
          JCLiteral literal = (JCLiteral)label.pat;
          int i = (Integer)literal.value + 1;
          
          if (i < count)
          {
            label = cases.get(i);
          }
          else
          {
            break;
          }
        }
        else
        {
          JCCase next = getLabel(getSetStateId(label.stats.head));
          
          if ((next != null) &&
            (isEmpty(label.stats.tail) || isContinue(label.stats.tail.head)))
          {
            label = next;
          }
          else
          {
            break;
          }
        }
      }
      
      entry.setValue(label);
      
      if (label.pat != null)
      {
        usedLabels.add(label);
      }
    }
    
    int id = 0;
    JCCase previous = null;
    
    for(int i = 0; i < count; ++i)
    {
      JCCase label = cases.get(i);
      
      if (!usedLabels.contains(label) && (previous != null))
      {
        JCStatement last = previous.stats.last();
        
        if (!isEmpty(label.stats) && 
          !isContinue(last) && 
          !(last instanceof JCReturn))
        {
          previous.stats = previous.stats.appendList(label.stats);
        }
        
        cases.set(i, null);
      
        continue;
      }

      JCLiteral literal = (JCLiteral)label.pat;
      
      literal.value = id++;
      
      if (previous == null)
      {
        previous = label;
        
        continue;
      }
      
      boolean found = false;
      List<JCStatement> list = previous.stats;
      
      while(!isEmpty(list))
      {
        JCStatement statement = list.head;
        
        if (found && isContinue(statement))
        {
          list.tail = null;
          list.head = null;
          
          break;
        }
        else
        {
          found = getLabel(getSetStateId(statement)) == label;
        }

        list = list.tail;
      }

      previous = label;
    }
    
    // Adjust label literals
    for(Map.Entry<JCLiteral, JCCase> entry: labelLiterals.entrySet())
    {
      JCCase label = entry.getValue();
      
      if (label != null)
      {
        JCLiteral literal = (JCLiteral)label.pat;
        
        entry.getKey().value = literal.value;
      }
    }
    
    
  }
  
  /**
   * Modifies original method.
   */
  private void modifyMethod()
  {
    methodTree.body = block(
      generateTypeAssert(),
      generateIteratorClass(),
      ret(newClass(ident(statePrefix))));
  }
  
  /**
   * Generates verification assert.
   * @return a verification assert.
   */
  private JCStatement generateTypeAssert()
  {
    return maker().Assert(
      maker().Binary(
        JCTree.EQ, 
        maker().TypeCast(
          maker().TypeApply(type("java.util.ArrayList"), list(itemType)), 
          maker().TypeCast(iteratorDecl.vartype, literal())), 
        literal()), 
      null);
  }

  /**
   * Generates iterator instance.
   * @return iterator instance.
   */
  private JCClassDecl generateIteratorClass()
  {
    ListBuffer<JCTree> members = ListBuffer.lb();
    
    members.appendList(classes);
    members.append(generateIterator());
    members.append(generateHasNext());
    members.append(generateNext());
    members.append(generateRemove());
    members.append(generateClose());
    members.append(generateMoveNext());
    
    if (createIterator)
    {
      members.append(generateIteratorIterator());
      members.append(generateObjectArrayIterator());
      members.append(generateArrayIterator());
    }

    for(JCVariableDecl variable: stateVariables)
    {
      members.append(variable);
    }
    
    return maker().ClassDef(
      maker().Modifiers(0), 
      name(statePrefix), 
      List.<JCTypeParameter>nil(), 
      null,
      list(
        maker().TypeApply(type("java.lang.Iterable"), list(copy(itemType))),
        maker().TypeApply(type("java.util.Iterator"), list(copy(itemType))),
        type("java.io.Closeable")), 
      members.toList());
  }

  /**
   * Generates iterator() method.
   * @return iterator() method.
   */
  private JCMethodDecl generateIterator()
  {
    return methodDef(
      Flags.PUBLIC, 
      name("iterator"),
      block(
        maker().If(
          maker().Binary(JCTree.EQ, ident(stateIdName), literal(0)), 
          block(
            assign(stateIdName, literal(1)), 
            ret(ident("this"))), 
          ret(newClass(ident(statePrefix))))), 
          maker().TypeApply(type("java.util.Iterator"), list(copy(itemType))));    
  }
  
  /**
   * Generates moveNext() method.
   * @return moveNext() method.
   */
  private JCMethodDecl generateMoveNext()
  {
    JCSwitch switchStatement = generateSwitch();
    JCStatement errorHandler = generateErrorHandler();

    return methodDef(
      Flags.PRIVATE,
      moveNextName,
      errorHandler != null ?
        block(
          varDef(0, errorName, type("java.lang.Throwable")),
          maker().WhileLoop(
            literal(true),
            block(
              maker().Try(
                block(switchStatement), 
                list(
                  maker().Catch(
                    varDef(0, name("e"), type("java.lang.Throwable")), 
                    block(assign(errorName, ident("e"))))),
                null),
              errorHandler))) :
        block(maker().WhileLoop(literal(true), switchStatement)),
      type(TypeTags.BOOLEAN));
  }

  /**
   * Generates iterator() method for an iterable.
   * @return iterator() method.
   */
  private JCMethodDecl generateIteratorIterator()
  {
    JCMethodDecl method = methodDef(
      Flags.PRIVATE,
      iteratorName,
      block(ret(invoke(ident("value"), name("iterator")))),
      type("java.util.Iterator"),
      varDef(Flags.FINAL, name("value"), type("java.lang.Iterable")));
    
    method.mods.annotations = list(suppressWarnings()); 

    return method;
  }

  /**
   * Generates iterator() method for an Object[].
   * @return iterator() method.
   */
  private JCMethodDecl generateObjectArrayIterator()
  {
    JCMethodDecl method = methodDef(
      Flags.PRIVATE,
      iteratorName,
      block(
        ret(
          invoke(
            invoke(
              ident("java.util.Arrays"), 
              name("asList"), 
              ident("value")),
            name("iterator")))),
      type("java.util.Iterator"),
      varDef(
        Flags.FINAL, 
        name("value"), 
        maker().TypeArray(type("java.lang.Object"))));
    
    method.mods.annotations = list(suppressWarnings()); 

    return method;
  }

  /**
   * Generates iterator() method for an array.
   * @return iterator() method.
   */
  private JCMethodDecl generateArrayIterator()
  {
    JCMethodDecl hasNextMethod = 
      methodDef(
        Flags.PUBLIC, 
        name("hasNext"), 
        block(ret(maker().Binary(JCTree.LT, ident("index"), ident("length")))), 
        type(TypeTags.BOOLEAN));  
      
    JCMethodDecl nextMethod = 
      methodDef(
        Flags.PUBLIC, 
        name("next"), 
        block(
          ret(
            invoke(
              type("java.lang.reflect.Array"),
              name("get"),
              ident("value"),
              maker().Unary(JCTree.POSTINC, ident("index"))))), 
        type("java.lang.Object"));  

    JCMethodDecl removeMethod = 
      methodDef(
        Flags.PUBLIC, 
        name("remove"), 
        block(
          maker().Throw(
            newClass(type("java.lang.UnsupportedOperationException")))), 
        type(TypeTags.VOID));

    JCVariableDecl indexVar = 
      varDef(Flags.PRIVATE, name("index"), type(TypeTags.INT));
    
    JCVariableDecl lengthVar =
      maker().VarDef(
        maker().Modifiers(Flags.FINAL), 
        name("length"), 
        type(TypeTags.INT),
        invoke(
          type("java.lang.reflect.Array"), 
          name("getLength"), 
          ident("value")));      
        
    JCMethodDecl method = methodDef(
      Flags.PRIVATE,
      iteratorName,
      block(
        lengthVar,
        ret(
          maker().NewClass(
            null, 
            List.<JCExpression>nil(), 
            type("java.util.Iterator"), 
            List.<JCExpression>nil(), 
            maker().ClassDef(
              maker().Modifiers(0), 
              empty, 
              List.<JCTypeParameter>nil(), 
              null, 
              List.<JCExpression>nil(), 
              list(
                hasNextMethod,
                nextMethod,
                removeMethod,
                indexVar))))),
      type("java.util.Iterator"),
      varDef(Flags.FINAL, name("value"), type("java.lang.Object")));
    
    method.mods.annotations = list(suppressWarnings()); 

    return method;
  }

  /**
   * Generates hasNext() method.
   * @return hasNext() method.
   */
  private JCMethodDecl generateHasNext()
  {
    return methodDef(
      Flags.PUBLIC,
      name("hasNext"),
      block(
        ifThen(
          maker().Unary(JCTree.NOT, ident(nextDefinedName)), 
          block(
            assign(hasNextName, invoke(null, moveNextName)),
            assign(nextDefinedName, literal(true)))),
        ret(ident(hasNextName))),
      type(TypeTags.BOOLEAN));
  }

  /**
   * Generates next() method.
   * @return next() method.
   */
  private JCMethodDecl generateNext()
  {
    return methodDef(
      Flags.PUBLIC,
      name("next"),
      block(
        ifThen(
          maker().Unary(JCTree.NOT, invoke(null, name("hasNext"))),
          maker().Throw(newClass(type("java.util.NoSuchElementException")))),
        assign(nextDefinedName, literal(false)),
        ret(ident(nextName))),
      copy(itemType));
  }

  /**
   * Generates remove() method.
   * @return remove() method.
   */
  private JCMethodDecl generateRemove()
  {
    return methodDef(
      Flags.PUBLIC,
      name("remove"),
      block(
        maker().Throw(
          newClass(type("java.lang.UnsupportedOperationException")))),
      type(TypeTags.VOID));
  }

  /**
   * Generates close() method.
   * @return close() method.
   */
  private JCMethodDecl generateClose()
  {
    JCStatement statement;
    
    if (breakCases.isEmpty())
    {
      statement = assign(stateIdName, literal(getBreakLabel(root)));
    }
    else
    {
      ListBuffer<JCCase> cases = ListBuffer.lb();
      Integer prev = null;
      
      for(JCCase breakCase: breakCases)
      {
        JCLiteral literal = (JCLiteral)breakCase.pat;
        Integer value = (Integer)literal.value;
        
        if ((prev == null) || ((int)prev != (int)value))
        {
          cases.append(breakCase);
          prev = value;
        }
      }
      
      cases.append(
        maker().Case(
          null, 
          list(
            assign(stateIdName, literal(getBreakLabel(root))),
            maker().Return(null))));
      
      statement =
        maker().DoLoop(
          maker().Switch(ident(stateIdName), cases.toList()), 
          invoke(null, moveNextName));
    }
    
    return methodDef(
      Flags.PUBLIC,
      name("close"),
      block(statement),
      type(TypeTags.VOID));
  }

  /**
   * Generates switch statements.
   * @return a switch statement.
   */
  private JCSwitch generateSwitch()
  {
    ListBuffer<JCCase> switchCases = ListBuffer.lb();

    for(JCCase statement: cases)
    {
      if (statement == null)
      {
        continue;
      }
      
      switchCases.append(statement);
    }
    
    switchCases.append(maker().Case(null, list(ret(literal(false)))));

    return maker().Switch(ident(stateIdName), switchCases.toList());
  }
  
  /**
   * Generates a switch for the error handlers.
   * @return a switch for the error handlers.
   */
  private JCStatement generateErrorHandler()
  {
    if (errorHandlers.isEmpty())
    {
      return null;
    }
    
    ListBuffer<JCCase> switchCases = ListBuffer.lb();
    HashSet<JCCase> labels = new HashSet<JCCase>();
    
    for(ErrorHandler handler: errorHandlers)
    {
      JCCase lastCase = null;
      
      for(int i = handler.begin; i < handler.end; ++i)
      {
        JCCase label = cases.get(i);
        
        if ((label != null) && labels.add(label))
        {
          lastCase = maker().Case(copy(label.pat), List.<JCStatement>nil());
          switchCases.append(lastCase);
        }
      }
      
      lastCase.stats = handler.statements.toList();
    }
    
    Name exception = name("e");
    JCExpression type = type("java.util.ConcurrentModificationException"); 
    
    switchCases.append(
      maker().Case(
        null, 
        list(
          assign(stateIdName, literal(getBreakLabel(root))),
          maker().VarDef(
            maker().Modifiers(0), 
            exception, 
            type, 
            newClass(type)),
            maker().Exec(
            invoke(ident(exception), name("initCause"), ident(errorName))),
            maker().Throw(ident(exception)))));

    return maker().Switch(ident(stateIdName), switchCases.toList());
  }

  /**
   * A processing environment.
   */
  private final JavacProcessingEnvironment environment;

  /**
   * A tree API.
   */
  private final Trees trees;
  
  /**
   * Type utils.
   */
  private final Types types;
  
  /**
   * Element utils.
   */
  private final Elements elements;
  
  /**
   * A name for "this".
   */
  private final Name _this;
  
  /**
   * A name for "super".
   */
  private final Name _super;
  
  /**
   * A name for "" (empty string).
   */
  private final Name empty;

  /**
   * A method.
   */
  private final ExecutableElement method;

  /**
   * A method three to validate.
   */
  final JCMethodDecl methodTree;
  
  /**
   * Item type.
   */
  private JCExpression itemType;
  
  /**
   * True if method is Iterable<T>, and false if Iterator<T>.
   */
  private boolean iterable;
  
  /**
   * Iterator declaration.
   */
  private JCVariableDecl iteratorDecl;
  
  /**
   * All names used.
   */
  private final HashSet<String> names = new HashSet<String>();
  
  /**
   * Local classes.
   */
  private final ListBuffer<JCTree> classes = ListBuffer.lb();

  /**
   * A list of yield scopes.
   */
  private ListBuffer<Scope> yields = ListBuffer.lb();
  
  /**
   * Break and continue statements.
   */
  private final ListBuffer<Scope> breaks = ListBuffer.lb();
  
  /**
   * Rewrite scopes.
   */
  private final ListBuffer<Scope> rewrites = ListBuffer.lb();

  /**
   * Variable declarations.
   */
  private final ListBuffer<Scope> variableDecls = ListBuffer.lb();

  /**
   * Variable declarations.
   */
  private ListBuffer<JCVariableDecl> stateVariables = ListBuffer.lb();

  /**
   * A root scope.
   */
  private Scope root;

  /**
   * Scope of the current node.
   */
  private Scope current;
  
  /**
   * All scopes.
   */
  private HashMap<JCTree, Scope> allScopes = new HashMap<JCTree, Scope>();

  /**
   * Refactored method body.
   */
  private ArrayList<JCCase> cases = new ArrayList<JCCase>();
  
  /**
   * A list buffer of current case.
   */
  private ListBuffer<JCStatement> statements = ListBuffer.lb();
  
  /**
   * Refactored method body.
   */
  private ListBuffer<JCCase> breakCases = ListBuffer.lb();
  
  /**
   * Refactored method exception handlers.
   */
  private ListBuffer<ErrorHandler> errorHandlers = ListBuffer.lb();

  /**
   * Finally exceptions. 
   */
  private int finallyBlocks;
  
  /**
   * Indicates that iterator method should be created.
   */
  private boolean createIterator;

  /**
   * Map of literals to labels.
   */
  private HashMap<JCLiteral, JCCase> labelLiterals = 
    new HashMap<JCLiteral, JCCase>();
  
  /**
   * Handler labels.
   */
  private HashSet<JCCase> usedLabels = new HashSet<JCCase>();

  /**
   * A tree maker.
   */
  private final TreeMaker maker;
  
  /**
   * A tree copier.
   */
  private final TreeCopier<JCTree> copier;

  /**
   * A state field prefix.
   */
  private String statePrefix;
  
  /**
   * A cache of names.
   */
  private HashMap<String, Name> namesCache = new HashMap<String, Name>();

  /**
   * State id name.
   */
  private Name stateIdName;

  /**
   * hasNext name.
   */
  private Name hasNextName;

  /**
   * nextDefined name.
   */
  private Name nextDefinedName;

  /**
   * Next value name.
   */
  private Name nextName;

  /**
   * Error name.
   */
  private Name errorName;

  /**
   * moveNext name.
   */
  private Name moveNextName;

  /**
   * iterator method name.
   */
  private Name iteratorName;

  /**
   * A source tree validator.
   */
  private class ValidationScanner extends TreeScanner
  {
    @Override
    public void visitBlock(final JCBlock tree)
    {
      current = new Scope(current, tree)
      {
        @Override
        public void refactor()
        {
          if (!isEmpty(tree.stats))
          {
            for(JCStatement statement: tree.stats)
            {
              refactorStatement(statement);
            }
          }

          addStatement(getBreakLabel(this));
        }
      };
      
      super.visitBlock(tree);
      current = current.parent;
    }
    
    @Override
    public void visitLabelled(final JCLabeledStatement tree)
    {
      current = new Scope(current, tree)
      {
        @Override
        public void refactor()
        {
          refactorStatement(tree.body);
        }
      };

      super.visitLabelled(tree);
      current = current.parent;
    }
    
    @Override
    public void visitForLoop(final JCForLoop tree)
    {
      current = new Scope(current, tree)
      {
        @Override
        public void refactor()
        {
          if (!isEmpty(tree.init))
          {
            for(JCStatement statement: tree.init)
            {
              refactorStatement(statement);
            }
          }
          
          JCCase label = label();
          JCCase breakLabel = getBreakLabel(this);

          addStatement(label);

          if ((tree.cond != null) && !isTrueLiteral(tree.cond))
          {
            addStatement(
              ifThen(
                maker().Unary(JCTree.NOT, copy(tree.cond)),
                block(
                  assign(stateIdName, literal(breakLabel)), 
                  doContinue())));
          }
          
          refactorStatement(tree.body);
          
          addStatement(getIterationLabel(this));
          
          if (!isEmpty(tree.step))
          {
            for(JCStatement statement: tree.step)
            {
              refactorStatement(statement);
            }
          }
          
          addStatement(assign(stateIdName, literal(label)));
          addStatement(doContinue());
          addStatement(breakLabel);
        }
      };
      
      super.visitForLoop(tree);
      current = current.parent;
    }

    @Override
    public void visitForeachLoop(final JCEnhancedForLoop tree)
    {
      current = new Scope(current, tree)
      {
        @Override
        public void refactor()
        {
          Name iteratorVarName = 
            name(statePrefix + "$iterator$" + tree.var.name);
          JCVariableDecl field = 
            varDef(Flags.PRIVATE, iteratorVarName, type("java.util.Iterator"));
          
          field.mods.annotations = list(suppressWarnings()); 
          
          stateVariables.append(field);
          
          addStatement(
            assign(
              iteratorVarName, 
              invoke(null, iteratorName, copy(tree.expr))));
          
          addStatement(getIterationLabel(this));

          addStatement(
            ifThen(
              maker().Unary(
                JCTree.NOT,  
                invoke(ident(iteratorVarName), name("hasNext"))), 
              block(
                assign(stateIdName, literal(getBreakLabel(this))),
                doContinue())));

          addStatement(
            assign(
              tree.var.name, 
              maker().TypeCast(
                copy(tree.var.vartype), 
                invoke(ident(iteratorVarName), name("next")))));
          
          refactorStatement(tree.body);
          addStatement(assign(stateIdName, literal(getIterationLabel(this))));
          addStatement(doContinue());
          addStatement(getBreakLabel(this));
        }
      };
      
      super.visitForeachLoop(tree);
      current = current.parent;
    }

    @Override
    public void visitDoLoop(final JCDoWhileLoop tree) 
    {
      current = new Scope(current, tree)
      {
        @Override
        public void refactor()
        {
          addStatement(getIterationLabel(this));
          
          refactorStatement(tree.body);

          addStatement(
            ifThen(
              copy(tree.cond), 
              block(
                assign(stateIdName, literal(getIterationLabel(this))),
                doContinue())));
          
          addStatement(getBreakLabel(this));
        }
      };
      
      super.visitDoLoop(tree);
      current = current.parent;
    }

    @Override
    public void visitWhileLoop(final JCWhileLoop tree) 
    {
      current = new Scope(current, tree)
      {
        @Override
        public void refactor()
        {
          addStatement(getIterationLabel(this));

          if (!isTrueLiteral(tree.cond))
          {
            addStatement(
              ifThen(
                maker().Unary(JCTree.NOT, copy(tree.cond)),
                block(
                  assign(stateIdName, literal(getBreakLabel(this))),
                  doContinue())));
          }
          
          refactorStatement(tree.body);
          
          addStatement(assign(stateIdName, literal(getIterationLabel(this))));
          addStatement(doContinue());
          addStatement(getBreakLabel(this));
        }
      };
      
      super.visitWhileLoop(tree);
      current = current.parent;
    }

    @Override
    public void visitIf(final JCIf tree) 
    {
      current = new Scope(current, tree)
      {
        @Override
        public void refactor()
        {
          JCCase label = tree.elsepart == null ? getBreakLabel(this) : label();
          
          if (!isTrueLiteral(tree.cond))
          {
            addStatement(
              ifThen(
                maker().Unary(JCTree.NOT, copy(tree.cond)),
                block(
                  assign(stateIdName, literal(label)), 
                  doContinue())));
          }

          if (tree.elsepart != null)
          {
            refactorStatement(tree.elsepart);
            addStatement(assign(stateIdName, literal(getBreakLabel(this))));
            addStatement(doContinue());
            addStatement(label);
          }
          
          refactorStatement(tree.thenpart);
          addStatement(getBreakLabel(this));
        }
      };
      
      super.visitIf(tree);
      current = current.parent;
    }

    @Override
    public void visitSwitch(final JCSwitch tree)
    {
      current = new Scope(current, tree)
      {
        @Override
        public void refactor()
        {
          JCCase breakLabel = getBreakLabel(this);
          JCSwitch switchStatement = 
            maker().Switch(copy(tree.selector), List.<JCCase>nil());
          
          addStatement(switchStatement);
          
          if (!isEmpty(tree.cases))
          {
            boolean hasDefault = false;
            ListBuffer<JCCase> cases = ListBuffer.lb();

            for(JCCase item: tree.cases)
            {
              if (item.pat == null)
              {
                hasDefault = true;
              }
              
              JCCase label = label();
              
              cases.append(
                maker().Case(
                  item.pat, 
                  list(
                    assign(stateIdName, literal(label)), 
                    doContinue())));

              addStatement(label);
              
              for(JCStatement statement: item.stats)
              {
                refactorStatement(statement);
              }
            }
            
            if (!hasDefault)
            {
              cases.append(
                maker().Case(
                  null, 
                  list(
                    assign(stateIdName, literal(breakLabel)), 
                    doContinue())));
            }

            switchStatement.cases = cases.toList();
          }
          
          addStatement(breakLabel);
        }
      };
      
      super.visitSwitch(tree);
      current = current.parent;
    }
    
    @Override
    public void visitTry(final JCTry tree)
    {
      current = new Scope(current, tree)
      {
        @Override
        public void refactor()
        {
          boolean hasFinally = tree.finalizer != null;
          boolean hasCatch = !isEmpty(tree.catchers);
          ErrorHandler catchHandler = null;
          ErrorHandler finallyHandler = null;
          JCCase tryLabel = label();
          JCCase finallyLabel;
          JCCase breakLabel = getBreakLabel(this);
          Name finallyErrorName = null;
          
          if (hasFinally)
          {
            finallyHandler = new ErrorHandler();
            finallyLabel = getFinallyLabel(this);
            
            ++finallyBlocks;
            finallyErrorName = name(statePrefix + "$exception" + finallyBlocks);
            labelName = name(statePrefix + "$id" + finallyBlocks);
            
            stateVariables.append(
              varDef(
                Flags.PRIVATE, 
                finallyErrorName, 
                type("java.lang.Throwable")));
            
            stateVariables.append(
              varDef(Flags.PRIVATE, labelName, type(TypeTags.INT)));

            addStatement(assign(finallyErrorName, literal()));
            addStatement(assign(labelName, literal(breakLabel)));
          }
          else
          {
            finallyLabel = breakLabel;
          }
          
          addStatement(assign(stateIdName, literal(tryLabel)));
          
          if (hasCatch)
          {
            catchHandler = new ErrorHandler();
            catchHandler.begin = cases.size();
          }
          else 
          {
            if (hasFinally)
            {
              finallyHandler.begin = cases.size();
            }
          }
          
          addStatement(tryLabel);
          refactorStatement(tree.body);
          addStatement(assign(stateIdName, literal(finallyLabel)));
          
          if (hasCatch)
          {
            addStatement(doContinue());
            catchHandler.end = cases.size();

            for(JCCatch catcher: tree.catchers)
            {
              JCCase label = label();
              
              usedLabels.add(label);
              addStatement(label);
              refactorStatement(catcher.body);
              addStatement(assign(stateIdName, literal(finallyLabel)));
              addStatement(doContinue());
              
              catchHandler.statements.append(
                ifThen(
                  maker().TypeTest(ident(errorName), copy(catcher.param.vartype)), 
                  block(
                    assign(
                      catcher.param.name, 
                      maker().TypeCast(
                        copy(catcher.param.vartype), 
                        ident(errorName))),
                    assign(stateIdName, literal(label)),
                    doContinue())));
            }

            errorHandlers.append(catchHandler);
            
            if (hasFinally)
            {
              finallyHandler.begin = catchHandler.end;
            }
          }
          
          if (hasFinally)
          {
            finallyHandler.end = cases.size();

            addStatement(finallyLabel);
            refactorStatement(tree.finalizer);
            
            addStatement(
              ifThen(
                maker().Binary(JCTree.NE, ident(finallyErrorName), literal()),
                block(
                  assign(errorName, ident(finallyErrorName)),
                  maker().Break(null))));

            Scope next = getFinallyScope(parent, null);
            
            if (next != null)
            {
              JCCase label = getFinallyLabel(next);
              
              addStatement(
                maker().If(
                  maker().Binary(JCTree.GT, ident(labelName), literal(label)), 
                  block(
                    assign(next.labelName, ident(labelName)),
                    assign(stateIdName, literal(label))), 
                  assign(stateIdName, ident(labelName))));
            }
            else
            {
              addStatement(assign(stateIdName, ident(labelName)));
            }
            
            addStatement(doContinue());

            finallyHandler.statements.append(
              assign(finallyErrorName, ident(errorName)));
            finallyHandler.statements.append(
              assign(stateIdName, literal(finallyLabel)));
            finallyHandler.statements.append(doContinue());
            
            usedLabels.add(finallyLabel);
            errorHandlers.append(finallyHandler);
          }
          
          addStatement(breakLabel);
        }
      };
      
      super.visitTry(tree);
      current = current.parent;
    }

    @Override
    public void visitVarDef(final JCVariableDecl tree)
    {
      if (tree == iteratorDecl)
      {
        return;
      }

      if (tree.name == iteratorDecl.name)
      {
        invalid("No variable with the same name as iterable is permitted.");
      }

      current = new Scope(current, tree)
      {
        @Override
        public void refactor()
        {
          if (tree.init != null)
          {
            addStatement(assign(tree.name, copy(tree.init)));
          }
        }
      };

      variableDecls.append(current);
      super.visitVarDef(tree);
      current = current.parent;
    }

    @Override 
    public void visitReturn(final JCReturn tree) 
    {
      if (!validateReturn(tree))
      {
        invalid(
          iterable ?
            "Return, if any, should be of the form: return items;" :
            "Return, if any, should be of the form: return items.iterate();");
      }
      
      current = new Scope(current, tree)
      {
        @Override
        public void refactor()
        {
          JCCase label = getBreakLabel(root);
          Scope next = getFinallyScope(parent, null);
          
          if (next == null)
          {
            addStatement(assign(stateIdName, literal(label)));
            addStatement(doContinue());
          }
          else
          {
            addStatement(assign(next.labelName, literal(label)));
            addStatement(assign(stateIdName, literal(getFinallyLabel(next))));
            addStatement(doContinue());
          }
        }
      };
      
      breaks.append(current);
      current = current.parent;
    }
    
    @Override
    public void visitBreak(final JCBreak tree)
    {
      Name label = tree.label;
      Scope target = null;
      
      if (label != null)
      {
        Scope labelScope = current;
        
        while(labelScope != null)
        {
          if (labelScope.node instanceof JCLabeledStatement)
          {
            JCLabeledStatement labeledStatement =
              (JCLabeledStatement)labelScope.node;
            
            if (label == labeledStatement.label)
            {
              if (target != null)
              {
                invalid("Invalid label.");  
              }
              
              target = labelScope;
            }
          }

          labelScope = labelScope.parent;
        }
      }
      else
      {
        Scope labelScope = current;
        
        while(labelScope != null)
        {
          if ((labelScope.node instanceof JCForLoop) ||
            (labelScope.node instanceof JCEnhancedForLoop) ||
            (labelScope.node instanceof JCWhileLoop) ||
            (labelScope.node instanceof JCDoWhileLoop) ||
            (labelScope.node instanceof JCSwitch))
          {
            target = labelScope;
            
            break;
          }

          labelScope = labelScope.parent;
        }
      }
      
      if (target == null)
      {
        invalid("Invalid break.");
      }
      
      current = new Scope(current, tree)
      {
        @Override
        public void refactor()
        {
          Scope next = getFinallyScope(parent, target);
          JCCase label = getBreakLabel(target);
          
          if (next == null)
          {
            addStatement(assign(stateIdName, literal(label)));
            addStatement(doContinue());
          }
          else
          {
            addStatement(assign(next.labelName, literal(label)));
            addStatement(assign(stateIdName, literal(getFinallyLabel(next))));
            addStatement(doContinue());
          }
        }
      };
      
      current.target = target;
      breaks.append(current);
      current = current.parent;
    }
    
    @Override
    public void visitContinue(final JCContinue tree)
    {
      Name label = tree.label;
      Scope target = null;
      
      if (label != null)
      {
        Scope labelScope = current;
        
        while(labelScope != null)
        {
          if (labelScope.node instanceof JCLabeledStatement)
          {
            JCLabeledStatement labeledStatement =
              (JCLabeledStatement)labelScope.node;
            
            if (label == labeledStatement.label)
            {
              if (target != null)
              {
                invalid("Invalid label.");  
              }
              
              if ((labelScope.node instanceof JCForLoop) ||
                (labelScope.node instanceof JCEnhancedForLoop) ||
                (labelScope.node instanceof JCWhileLoop) ||
                (labelScope.node instanceof JCDoWhileLoop))
              {
                target = labelScope;
              }
              else
              {
                invalid("Invalid continue.");
              }
            }
          }
          
          labelScope = labelScope.parent;
        }
      }
      else
      {
        Scope labelScope = current;
        
        while(labelScope != null)
        {
          if ((labelScope.node instanceof JCForLoop) ||
            (labelScope.node instanceof JCEnhancedForLoop) ||
            (labelScope.node instanceof JCWhileLoop) ||
            (labelScope.node instanceof JCDoWhileLoop))
          {
            target = labelScope;
            
            break;
          }

          labelScope = labelScope.parent;
        }
      }
      
      if (target == null)
      {
        invalid("Invalid continue.");
      }
      
      current = new Scope(current, tree)
      {
        @Override
        public void refactor()
        {
          Scope next = getFinallyScope(parent, target);
          JCCase label = getIterationLabel(target);

          if (next == null)
          {
            addStatement(assign(stateIdName, literal(label)));
            addStatement(doContinue());
          }
          else
          {
            addStatement(assign(next.labelName, literal(label)));
            addStatement(assign(stateIdName, literal(getFinallyLabel(next))));
            addStatement(doContinue());
          }
        }
      };
        
      current.target = target;
      breaks.append(current);
      current = current.parent;
    }
    
    @Override
    public void visitApply(JCMethodInvocation tree)
    {
      if (tree.meth instanceof JCIdent)
      {
        Name name = TreeInfo.fullName(tree.meth);
        
        if ((iterable && (name == name("iterator"))) ||
          (name == name("hasNext")) ||
          (name == name("next")) ||
          (name == name("remove")) ||
          (name == name("close")))
        {
          invalid("Cannot call method " + name + "(), as it is hidden.");
        }
      }
      
      super.visitApply(tree);
    }

    @Override
    public void visitExec(JCExpressionStatement tree)
    {
      final JCExpression expression = getYieldExpression(tree.expr);
      
      if (expression != null)
      {
        current = new Scope(current, tree)
        {
          @Override
          public void refactor()
          {
            JCCase label = getBreakLabel(this);
            
            addStatement(assign(nextName, copy(expression)));
            addStatement(assign(stateIdName, literal(label)));
            addStatement(ret(literal(true)));
            addStatement(label);
            
            Scope next = getFinallyScope(parent, null);
            
            if (next != null)
            {
              breakCases.append(
                maker().Case(
                  literal(label), 
                  list(
                    assign(next.labelName, literal(getBreakLabel(root))),
                    assign(stateIdName, literal(getFinallyLabel(next))),
                    doContinue())));
            }
          }
        };
        
        yields.append(current);
        scan(expression);
        current = current.parent;
      }
      else
      {
        super.visitExec(tree);
      }
    }
    
    @Override
    public void visitIdent(JCIdent tree)
    {
      if (tree.name == iteratorDecl.name)
      {
        invalid("No use of iterable is permitted in contexts other " +
          "than defined by the yield pattern.");
      }
      
      if (tree.name == _this)
      {
        invalid("No unqualified \"this\" expression is permitted.");
      }

      if (tree.name == _super)
      {
        invalid("No unqualified \"super\" expression is permitted.");
      }
      
      names.add(tree.name.toString());
      
      super.visitIdent(tree);
    }

    @Override
    public void visitNewClass(JCNewClass tree)
    { 
      scan(tree.encl);
      scan(tree.clazz);
      scan(tree.args);
    }

    @Override
    public void visitClassDef(JCClassDecl tree)
    { 
      classes.add(tree);
    }
  }
}